#!/usr/bin/env python
"""Markdown formatting and linting utility for the project.

This script provides automated Markdown formatting using markdownlint,
which uses the existing .markdownlint.json configuration.

Usage:
    uv run format-md                    # Format all Markdown files
    uv run format-md --check            # Check formatting only, don't modify
    uv run format-md README.md          # Format specific files
    uv run format-md docs/              # Format all files in directory

    # Or directly:
    python utils/format_markdown.py
"""

import argparse
import shutil
import subprocess
import sys
from pathlib import Path


def run_command(cmd: list[str], check: bool = True) -> subprocess.CompletedProcess:
    """Run a command and optionally exit on failure."""
    print(f"Running: {' '.join(cmd)}")
    result = subprocess.run(cmd, capture_output=True, text=True)

    if check and result.returncode != 0:
        if result.stdout:
            print(result.stdout)
        if result.stderr:
            print(result.stderr)
        sys.exit(result.returncode)

    return result


def check_markdown_tool() -> bool:
    """Check if markdownlint is installed."""
    if not shutil.which("markdownlint"):
        print("❌ markdownlint is not installed!")
        print("\nTo install:")
        print("  npm install -g markdownlint-cli")
        print("  # or")
        print("  brew install markdownlint-cli")
        return False
    return True


def get_markdown_files(paths: list[str] | None = None) -> list[Path]:
    """Get all Markdown files to format."""
    if not paths:
        # Default: format all .md files in the project
        paths = ["."]

    md_files = []
    for path_str in paths:
        path = Path(path_str)
        if path.is_file():
            if path.suffix.lower() in {".md", ".markdown"}:
                md_files.append(path)
        elif path.is_dir():
            # Recursively find all .md files
            md_files.extend(path.glob("**/*.md"))
            md_files.extend(path.glob("**/*.markdown"))

    # Filter out common directories to ignore
    ignore_dirs = {".venv", "venv", "node_modules", ".git", "dist", "build", ".tox"}
    md_files = [f for f in md_files if not any(ignored in f.parts for ignored in ignore_dirs)]

    return sorted(set(md_files))


def format_markdown(files: list[Path], check_only: bool = False) -> bool:
    """Format Markdown files using markdownlint."""
    if not files:
        print("No Markdown files found to format.")
        return True

    print(f"\n{'🔍 Checking' if check_only else '🔧 Formatting'} {len(files)} Markdown file(s)...")

    # Build markdownlint command
    cmd = ["markdownlint"]

    # Use config file if it exists
    config_file = Path(".markdownlint.json")
    if config_file.exists():
        cmd.extend(["--config", str(config_file)])

    if not check_only:
        cmd.append("--fix")

    # Add file paths
    cmd.extend(str(f) for f in files)

    # Run markdownlint
    result = run_command(cmd, check=False)

    if result.returncode != 0:
        if check_only:
            print("❌ Some files need formatting")
            if result.stdout:
                print(result.stdout)
            print("\n🔧 To fix formatting issues, run:")
            print("  uv run format-md")
        else:
            print("❌ Formatting failed")
            if result.stdout:
                print(result.stdout)
            if result.stderr:
                print(result.stderr)
        return False

    if check_only:
        print("✅ All Markdown files are properly formatted!")
    else:
        print("✅ Markdown formatting complete!")

    return True


def main() -> None:
    """Main entry point for the formatter."""
    parser = argparse.ArgumentParser(description="Format and lint Markdown files with markdownlint")
    parser.add_argument("paths", nargs="*", help="Files or directories to format (default: all .md files)")
    parser.add_argument("--check", action="store_true", help="Check formatting only, don't modify files")

    args = parser.parse_args()

    # Check if markdownlint is installed
    if not check_markdown_tool():
        sys.exit(1)

    # Get files to format
    md_files = get_markdown_files(args.paths)

    if not md_files:
        print("No Markdown files found.")
        sys.exit(0)

    # Format the files
    success = format_markdown(md_files, check_only=args.check)

    if not success:
        sys.exit(1)


if __name__ == "__main__":
    main()



"""Add strategy and strategy_preset tables

Revision ID: 0e6547a3ccf0
Revises: c1ceaae14e9f
Create Date: 2025-08-05 08:20:13.498194

"""

from collections.abc import Sequence

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "0e6547a3ccf0"
down_revision: str | Sequence[str] | None = "c1ceaae14e9f"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "strategies",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("class_name", sa.String(), nullable=False),
        sa.Column("module_path", sa.String(), nullable=False),
        sa.Column("description", sa.String(), nullable=True),
        sa.Column("category", sa.String(), nullable=True),
        sa.Column("is_active", sa.Boolean(), nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=False),
        sa.Column("updated_at", sa.DateTime(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_strategies_name"), "strategies", ["name"], unique=True)
    op.create_table(
        "strategy_presets",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("strategy_id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("parameters_json", sa.String(), nullable=False),
        sa.Column("description", sa.String(), nullable=True),
        sa.Column("is_default", sa.Boolean(), nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=False),
        sa.Column("updated_at", sa.DateTime(), nullable=False),
        sa.ForeignKeyConstraint(
            ["strategy_id"],
            ["strategies.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("strategy_id", "name", name="unique_strategy_preset"),
    )
    op.create_index(op.f("ix_strategy_presets_name"), "strategy_presets", ["name"], unique=False)
    op.create_index(op.f("ix_strategy_presets_strategy_id"), "strategy_presets", ["strategy_id"], unique=False)
    op.drop_index(op.f("idx_dividends_date"), table_name="dividends")
    op.drop_index(op.f("idx_options_calls_symbol_exp_strike"), table_name="options_calls")
    op.drop_index(op.f("idx_options_puts_symbol_exp_strike"), table_name="options_puts")
    op.drop_index(op.f("idx_price_history_date"), table_name="price_history")
    op.drop_index(op.f("idx_price_history_symbol_date_interval"), table_name="price_history")
    op.drop_index(op.f("idx_price_history_symbol_interval"), table_name="price_history")
    op.drop_index(op.f("idx_splits_date"), table_name="splits")
    op.drop_index(op.f("idx_stock_info_symbol"), table_name="stock_info")
    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_index(op.f("idx_stock_info_symbol"), "stock_info", ["symbol"], unique=False)
    op.create_index(op.f("idx_splits_date"), "splits", ["date"], unique=False)
    op.create_index(op.f("idx_price_history_symbol_interval"), "price_history", ["symbol", "interval"], unique=False)
    op.create_index(
        op.f("idx_price_history_symbol_date_interval"), "price_history", ["symbol", "date", "interval"], unique=False
    )
    op.create_index(op.f("idx_price_history_date"), "price_history", ["date"], unique=False)
    op.create_index(
        op.f("idx_options_puts_symbol_exp_strike"),
        "options_puts",
        ["symbol", "expiration_date", "strike"],
        unique=False,
    )
    op.create_index(
        op.f("idx_options_calls_symbol_exp_strike"),
        "options_calls",
        ["symbol", "expiration_date", "strike"],
        unique=False,
    )
    op.create_index(op.f("idx_dividends_date"), "dividends", ["date"], unique=False)
    op.drop_index(op.f("ix_strategy_presets_strategy_id"), table_name="strategy_presets")
    op.drop_index(op.f("ix_strategy_presets_name"), table_name="strategy_presets")
    op.drop_table("strategy_presets")
    op.drop_index(op.f("ix_strategies_name"), table_name="strategies")
    op.drop_table("strategies")
    # ### end Alembic commands ###



"""Tests for backtesting module."""

from datetime import datetime
from unittest.mock import Mock

import numpy as np
import pandas as pd
import pytest

from stockula.backtesting import (
    BacktestRunner,
    DoubleEMACrossStrategy,
    MACDStrategy,
    RSIStrategy,
    SMACrossStrategy,
    TRIMACrossStrategy,
    TripleEMACrossStrategy,
)
from stockula.config.models import (
    BacktestResult,
    BrokerConfig,
    PortfolioBacktestResults,
    StrategyBacktestSummary,
)


class TestBacktestRunner:
    """Test BacktestRunner class."""

    def test_initialization(self):
        """Test BacktestRunner initialization."""
        runner = BacktestRunner(cash=50000, commission=0.001, margin=2.0)
        assert runner.cash == 50000
        assert runner.commission == 0.001
        assert runner.margin == 2.0

        # Test defaults
        runner_default = BacktestRunner()
        assert runner_default.cash == 10000
        assert runner_default.commission == 0.002
        assert runner_default.margin == 1.0

    def test_run_with_data(self, backtest_data):
        """Test running backtest with data."""
        runner = BacktestRunner()
        results = runner.run(backtest_data, SMACrossStrategy)

        # Check result structure - backtesting library returns a Series
        assert isinstance(results, pd.Series)
        assert "Return [%]" in results.index
        assert "Sharpe Ratio" in results.index
        assert "Max. Drawdown [%]" in results.index
        assert "# Trades" in results.index

        # Results should be numeric
        assert isinstance(results["Return [%]"], int | float)
        assert isinstance(results["# Trades"], int)

    def test_run_from_symbol(self, mock_data_fetcher, backtest_data):
        """Test running backtest from symbol."""
        # Create a mock that returns our backtest data
        mock_fetcher = Mock()
        mock_fetcher.get_stock_data = Mock(return_value=backtest_data)

        # Create runner with mocked data fetcher
        runner = BacktestRunner(data_fetcher=mock_fetcher)

        results = runner.run_from_symbol("AAPL", SMACrossStrategy)

        assert isinstance(results, pd.Series)  # backtesting library returns Series
        assert "Return [%]" in results.index
        mock_fetcher.get_stock_data.assert_called_once()

    def test_optimize(self, backtest_data):
        """Test strategy optimization."""
        runner = BacktestRunner()

        # Define parameter ranges
        param_ranges = {
            "fast_period": range(5, 15, 5),  # [5, 10]
            "slow_period": range(20, 30, 5),  # [20, 25]
        }

        optimal_results = runner.optimize(backtest_data, SMACrossStrategy, **param_ranges)

        # Check result - optimize returns a Series with the best backtest result
        assert isinstance(optimal_results, pd.Series)
        assert "Return [%]" in optimal_results.index

        # Check that the strategy has the optimized parameters
        assert hasattr(optimal_results._strategy, "fast_period")
        assert hasattr(optimal_results._strategy, "slow_period")
        assert optimal_results._strategy.fast_period in param_ranges["fast_period"]
        assert optimal_results._strategy.slow_period in param_ranges["slow_period"]

    def test_invalid_strategy(self, backtest_data):
        """Test error handling for invalid strategy."""
        runner = BacktestRunner()

        class InvalidStrategy:
            """Strategy without required methods."""

            pass

        with pytest.raises(TypeError, match="must be a Strategy sub-type"):
            runner.run(backtest_data, InvalidStrategy)


class TestSMACrossStrategy:
    """Test SMA Crossover Strategy."""

    def test_strategy_attributes(self):
        """Test strategy has required attributes."""
        assert hasattr(SMACrossStrategy, "fast_period")
        assert hasattr(SMACrossStrategy, "slow_period")
        assert SMACrossStrategy.fast_period < SMACrossStrategy.slow_period

    def test_strategy_execution(self, backtest_data):
        """Test strategy execution."""
        runner = BacktestRunner()
        results = runner.run(backtest_data, SMACrossStrategy)

        # Check that results are valid (strategy might not trade with default parameters)
        assert isinstance(results, pd.Series)
        assert "Return [%]" in results.index
        assert "# Trades" in results.index
        assert results["# Trades"] >= 0  # May be 0 if no signals generated

    def test_custom_parameters(self, backtest_data):
        """Test strategy with custom parameters."""

        # Create custom strategy class
        class CustomSMACross(SMACrossStrategy):
            fast_period = 5
            slow_period = 15

        runner = BacktestRunner()
        results = runner.run(backtest_data, CustomSMACross)

        # More sensitive parameters should generate more trades
        assert results["# Trades"] >= 0


class TestRSIStrategy:
    """Test RSI Strategy."""

    def test_strategy_attributes(self):
        """Test strategy has required attributes."""
        assert hasattr(RSIStrategy, "rsi_period")
        assert hasattr(RSIStrategy, "oversold_threshold")
        assert hasattr(RSIStrategy, "overbought_threshold")
        assert RSIStrategy.oversold_threshold < RSIStrategy.overbought_threshold

    def test_strategy_execution(self, backtest_data):
        """Test strategy execution."""
        runner = BacktestRunner()
        results = runner.run(backtest_data, RSIStrategy)

        # Check results
        assert isinstance(results["Return [%]"], int | float)
        assert isinstance(results["# Trades"], int)


class TestMACDStrategy:
    """Test MACD Strategy."""

    def test_strategy_attributes(self):
        """Test strategy has required attributes."""
        assert hasattr(MACDStrategy, "fast_period")
        assert hasattr(MACDStrategy, "slow_period")
        assert hasattr(MACDStrategy, "signal_period")

    def test_strategy_execution(self, backtest_data):
        """Test strategy execution."""
        runner = BacktestRunner()
        results = runner.run(backtest_data, MACDStrategy)

        # Check results
        assert isinstance(results["Return [%]"], int | float)
        assert results["Max. Drawdown [%]"] <= 0  # Drawdown is negative


class TestDoubleEMACrossStrategy:
    """Test Double EMA Cross Strategy."""

    def test_strategy_attributes(self):
        """Test strategy has required attributes."""
        assert hasattr(DoubleEMACrossStrategy, "fast_period")
        assert hasattr(DoubleEMACrossStrategy, "slow_period")
        assert hasattr(DoubleEMACrossStrategy, "momentum_atr_multiple")
        assert hasattr(DoubleEMACrossStrategy, "speculative_atr_multiple")

    def test_minimum_data_requirements(self):
        """Test minimum data requirements."""
        min_days = DoubleEMACrossStrategy.get_min_required_days()
        assert min_days == 75  # 55 + 20 buffer

        # Test with exact date
        end_date = "2024-01-01"
        start_date = DoubleEMACrossStrategy.get_recommended_start_date(end_date)

        # Should be at least min_days before end date
        days_diff = (datetime.strptime(end_date, "%Y-%m-%d") - datetime.strptime(start_date, "%Y-%m-%d")).days
        assert days_diff >= min_days

    def test_insufficient_data_warning(self):
        """Test warning for insufficient data."""
        # Create data with insufficient history
        short_data = pd.DataFrame(
            {
                "Open": [100] * 50,
                "High": [101] * 50,
                "Low": [99] * 50,
                "Close": [100] * 50,
                "Volume": [1000000] * 50,
            },
            index=pd.date_range("2023-01-01", periods=50),
        )

        runner = BacktestRunner()

        # The strategy should issue a warning about insufficient data
        import warnings

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            runner.run(short_data, DoubleEMACrossStrategy)

            # Check that a warning was issued
            assert len(w) > 0
            assert any("Insufficient data" in str(warning.message) for warning in w)

    def test_strategy_with_categories(self, backtest_data):
        """Test strategy with different ATR multiples for categories."""

        # Modify strategy to test category-specific behavior
        class TestDoubleEMA(DoubleEMACrossStrategy):
            momentum_atr_multiple = 1.5
            speculative_atr_multiple = 1.0

        runner = BacktestRunner()
        results = runner.run(backtest_data, TestDoubleEMA)

        assert isinstance(results, pd.Series)
        assert "Return [%]" in results.index


class TestTripleEMACrossStrategy:
    """Test Triple EMA Cross Strategy."""

    def test_strategy_attributes(self):
        """Test strategy has required attributes."""
        assert hasattr(TripleEMACrossStrategy, "fast_period")
        assert hasattr(TripleEMACrossStrategy, "slow_period")
        assert hasattr(TripleEMACrossStrategy, "atr_period")
        assert hasattr(TripleEMACrossStrategy, "atr_multiple")

    def test_minimum_data_requirements(self):
        """Test minimum data requirements."""
        min_days = TripleEMACrossStrategy.get_min_required_days()
        assert min_days == 81  # 3*21-2=61 + 20 buffer

    def test_tema_calculation(self, backtest_data):
        """Test TEMA calculation is different from regular EMA."""
        runner = BacktestRunner()

        # Run strategy
        results = runner.run(backtest_data, TripleEMACrossStrategy)

        # TEMA should produce different results than regular EMA cross
        ema_results = runner.run(backtest_data, SMACrossStrategy)

        # Results should differ (different strategies)
        assert results["# Trades"] != ema_results["# Trades"]


class TestTRIMACrossStrategy:
    """Test TRIMA Cross Strategy."""

    def test_strategy_attributes(self):
        """Test strategy has required attributes."""
        assert hasattr(TRIMACrossStrategy, "fast_period")
        assert hasattr(TRIMACrossStrategy, "slow_period")
        assert hasattr(TRIMACrossStrategy, "atr_period")
        assert hasattr(TRIMACrossStrategy, "atr_multiple")

    def test_minimum_data_requirements(self):
        """Test minimum data requirements."""
        min_days = TRIMACrossStrategy.get_min_required_days()
        assert min_days == 76  # 2*28=56 + 20 buffer

    def test_strategy_smoothing(self, backtest_data):
        """Test that TRIMA provides smoother signals."""
        runner = BacktestRunner()

        # Add some noise to data
        noisy_data = backtest_data.copy()
        noise = np.random.normal(0, 2, len(noisy_data))
        noisy_data["Close"] = noisy_data["Close"] + noise
        noisy_data["High"] = noisy_data["High"] + abs(noise)
        noisy_data["Low"] = noisy_data["Low"] - abs(noise)

        # Run TRIMA strategy
        trima_results = runner.run(noisy_data, TRIMACrossStrategy)

        # Run regular SMA strategy for comparison
        sma_results = runner.run(noisy_data, SMACrossStrategy)

        # TRIMA should produce fewer trades due to smoothing
        assert trima_results["# Trades"] <= sma_results["# Trades"] * 1.5


class TestBacktestingEdgeCases:
    """Test edge cases in backtesting."""

    def test_no_trades(self):
        """Test strategy that generates no trades."""
        # Create flat data
        flat_data = pd.DataFrame(
            {
                "Open": [100] * 100,
                "High": [100] * 100,
                "Low": [100] * 100,
                "Close": [100] * 100,
                "Volume": [1000000] * 100,
            },
            index=pd.date_range("2023-01-01", periods=100),
        )

        runner = BacktestRunner()
        results = runner.run(flat_data, SMACrossStrategy)

        assert results["# Trades"] == 0
        assert results["Return [%]"] == 0.0

    def test_single_data_point(self):
        """Test with single data point."""
        single_point = pd.DataFrame(
            {
                "Open": [100],
                "High": [101],
                "Low": [99],
                "Close": [100],
                "Volume": [1000000],
            },
            index=[datetime.now()],
        )

        runner = BacktestRunner()

        # Should handle gracefully
        results = runner.run(single_point, SMACrossStrategy)
        assert results["# Trades"] == 0

    def test_extreme_volatility(self):
        """Test with extreme volatility."""
        dates = pd.date_range("2023-01-01", periods=100)

        # Create extreme volatility
        prices = []
        for i in range(len(dates)):
            if i % 2 == 0:
                prices.append(100)
            else:
                prices.append(150)  # 50% swings

        volatile_data = pd.DataFrame(
            {
                "Open": prices,
                "High": [p * 1.1 for p in prices],
                "Low": [p * 0.9 for p in prices],
                "Close": prices,
                "Volume": [1000000] * len(dates),
            },
            index=dates,
        )

        runner = BacktestRunner()
        results = runner.run(volatile_data, RSIStrategy)

        # Should handle without errors
        assert isinstance(results["Return [%]"], int | float)
        assert isinstance(results["Max. Drawdown [%]"], int | float)

    def test_commission_impact(self, backtest_data):
        """Test impact of commission on results."""
        # Run with no commission
        runner_no_comm = BacktestRunner(commission=0.0)
        results_no_comm = runner_no_comm.run(backtest_data, SMACrossStrategy)

        # Run with high commission
        runner_high_comm = BacktestRunner(commission=0.01)  # 1%
        results_high_comm = runner_high_comm.run(backtest_data, SMACrossStrategy)

        # High commission should reduce returns
        if results_no_comm["# Trades"] > 0:
            assert results_high_comm["Return [%]"] < results_no_comm["Return [%]"]

    def test_runner_with_broker_config(self, backtest_data):
        """Test BacktestRunner with broker configuration."""
        # Test with Robinhood preset
        broker_config = BrokerConfig.from_broker_preset("robinhood")
        runner = BacktestRunner(broker_config=broker_config)

        assert runner.broker_config == broker_config
        assert callable(runner.commission)

        # Run backtest
        results = runner.run(backtest_data, SMACrossStrategy)
        assert isinstance(results, pd.Series)
        assert "Return [%]" in results.index

    def test_convert_results_to_backtest_result(self, backtest_data):
        """Test converting raw backtest results to BacktestResult model."""
        runner = BacktestRunner()
        results = runner.run(backtest_data, SMACrossStrategy)

        # Convert to BacktestResult
        backtest_result = BacktestResult(
            ticker="TEST",
            strategy="SMACross",
            parameters={"fast_period": 10, "slow_period": 20},
            return_pct=results["Return [%]"],
            sharpe_ratio=results["Sharpe Ratio"],
            max_drawdown_pct=results["Max. Drawdown [%]"],
            num_trades=results["# Trades"],
            win_rate=results.get("Win Rate [%]") if results["# Trades"] > 0 else None,
        )

        assert backtest_result.ticker == "TEST"
        assert backtest_result.strategy == "SMACross"
        assert isinstance(backtest_result.return_pct, int | float)
        assert isinstance(backtest_result.num_trades, int)


class TestBacktestDataStructures:
    """Test the new backtest data structures in integration scenarios."""

    def test_full_portfolio_backtest_workflow(self, backtest_data):
        """Test complete workflow with new data structures."""
        # Setup portfolio with multiple strategies
        strategies = [
            ("SMACross", SMACrossStrategy, {}),
            ("RSI", RSIStrategy, {"period": 14}),
        ]

        strategy_summaries = []

        for strategy_name, strategy_class, params in strategies:
            runner = BacktestRunner(cash=10000)
            results = runner.run(backtest_data, strategy_class)

            # Create BacktestResult
            backtest_result = BacktestResult(
                ticker="TEST",
                strategy=strategy_name,
                parameters=params,
                return_pct=results["Return [%]"],
                sharpe_ratio=results["Sharpe Ratio"],
                max_drawdown_pct=results["Max. Drawdown [%]"],
                num_trades=results["# Trades"],
                win_rate=results.get("Win Rate [%]") if results["# Trades"] > 0 else None,
            )

            # Create strategy summary
            summary = StrategyBacktestSummary(
                strategy_name=strategy_name,
                parameters=params,
                initial_portfolio_value=10000.0,
                final_portfolio_value=10000.0 * (1 + backtest_result.return_pct / 100),
                total_return_pct=backtest_result.return_pct,
                total_trades=backtest_result.num_trades,
                winning_stocks=1 if backtest_result.return_pct > 0 else 0,
                losing_stocks=1 if backtest_result.return_pct < 0 else 0,
                average_return_pct=backtest_result.return_pct,
                average_sharpe_ratio=backtest_result.sharpe_ratio,
                detailed_results=[backtest_result],
            )

            strategy_summaries.append(summary)

        # Create portfolio results
        portfolio_results = PortfolioBacktestResults(
            initial_portfolio_value=10000.0,
            initial_capital=10000.0,
            date_range={"start": "2023-01-01", "end": "2023-12-31"},
            broker_config={
                "name": "robinhood",
                "commission_type": "fixed",
                "commission_value": 0.0,
            },
            strategy_summaries=strategy_summaries,
        )

        # Verify structure
        assert len(portfolio_results.strategy_summaries) == 2
        assert portfolio_results.strategy_summaries[0].strategy_name == "SMACross"
        assert portfolio_results.strategy_summaries[1].strategy_name == "RSI"

        # Test serialization
        data = portfolio_results.model_dump()
        assert isinstance(data, dict)
        assert "strategy_summaries" in data
        assert len(data["strategy_summaries"]) == 2

    def test_broker_config_impact_on_results(self, backtest_data):
        """Test how different broker configs affect results."""
        brokers = ["robinhood", "interactive_brokers", "fidelity"]
        results_by_broker = {}

        for broker_name in brokers:
            broker_config = BrokerConfig.from_broker_preset(broker_name)
            runner = BacktestRunner(cash=10000, broker_config=broker_config)
            results = runner.run(backtest_data, SMACrossStrategy)

            results_by_broker[broker_name] = BacktestResult(
                ticker="TEST",
                strategy="SMACross",
                parameters={},
                return_pct=results["Return [%]"],
                sharpe_ratio=results["Sharpe Ratio"],
                max_drawdown_pct=results["Max. Drawdown [%]"],
                num_trades=results["# Trades"],
                win_rate=results.get("Win Rate [%]") if results["# Trades"] > 0 else None,
            )

        # Verify different brokers may have different results due to fees
        # Interactive Brokers has per-share fees, others are zero commission
        if results_by_broker["robinhood"].num_trades > 0:
            # IB should have lower returns due to commission
            assert results_by_broker["interactive_brokers"].return_pct <= results_by_broker["robinhood"].return_pct



"""Forecasting module with AutoGluon and Chronos backends (falls back to simple)."""

from .backends import (
    AUTOGLUON_AVAILABLE,
    CHRONOS_AVAILABLE,
    AutoGluonBackend,
    ChronosBackend,
    ForecastBackend,
    ForecastResult,
    SimpleForecastBackend,
)
from .factory import create_forecast_backend
from .manager import ForecastingManager

__all__ = [
    # Core components
    "ForecastingManager",
    # Backend abstraction
    "ForecastBackend",
    "ForecastResult",
    "AutoGluonBackend",
    "SimpleForecastBackend",
    "ChronosBackend",
    "create_forecast_backend",
    "AUTOGLUON_AVAILABLE",
    "CHRONOS_AVAILABLE",
]



"""Tests for interface definitions in stockula.interfaces."""

import inspect
from abc import ABC
from unittest.mock import Mock

import pandas as pd

from stockula.interfaces import (
    IBacktestRunner,
    IDatabaseManager,
    IDataFetcher,
    IDomainFactory,
    ILoggingManager,
    IStockForecaster,
    ITechnicalIndicators,
)


class TestIDataFetcher:
    """Test IDataFetcher interface definition."""

    def test_is_abstract_base_class(self):
        """Test that IDataFetcher is an abstract base class."""
        assert issubclass(IDataFetcher, ABC)
        assert hasattr(IDataFetcher, "__abstractmethods__")

    def test_has_required_methods(self):
        """Test that IDataFetcher has all required abstract methods."""
        required_methods = ["get_stock_data", "get_current_prices", "get_info"]

        for method in required_methods:
            assert hasattr(IDataFetcher, method)
            assert callable(getattr(IDataFetcher, method))

    def test_get_stock_data_signature(self):
        """Test get_stock_data method signature."""
        method = IDataFetcher.get_stock_data
        sig = inspect.signature(method)

        # Check parameter names
        param_names = list(sig.parameters.keys())
        expected_params = [
            "self",
            "symbol",
            "start",
            "end",
            "interval",
            "force_refresh",
        ]
        assert param_names == expected_params

    def test_get_current_prices_signature(self):
        """Test get_current_prices method signature."""
        method = IDataFetcher.get_current_prices
        sig = inspect.signature(method)

        param_names = list(sig.parameters.keys())
        assert "self" in param_names
        assert "symbols" in param_names
        assert "show_progress" in param_names

    def test_cannot_instantiate_directly(self):
        """Test that IDataFetcher cannot be instantiated directly."""
        try:
            IDataFetcher()
            raise AssertionError("Should not be able to instantiate abstract class")
        except TypeError:
            pass  # Expected


class TestIDatabaseManager:
    """Test IDatabaseManager interface definition."""

    def test_is_abstract_base_class(self):
        """Test that IDatabaseManager is an abstract base class."""
        assert issubclass(IDatabaseManager, ABC)

    def test_has_required_methods(self):
        """Test that IDatabaseManager has all required abstract methods."""
        required_methods = ["get_price_history", "store_price_history", "has_data"]

        for method in required_methods:
            assert hasattr(IDatabaseManager, method)

    def test_cannot_instantiate_directly(self):
        """Test that IDatabaseManager cannot be instantiated directly."""
        try:
            IDatabaseManager()
            raise AssertionError("Should not be able to instantiate abstract class")
        except TypeError:
            pass


class TestIStockForecaster:
    """Test IStockForecaster interface definition."""

    def test_is_abstract_base_class(self):
        """Test that IStockForecaster is an abstract base class."""
        assert issubclass(IStockForecaster, ABC)

    def test_has_required_methods(self):
        """Test that IStockForecaster has all required abstract methods."""
        required_methods = ["forecast", "forecast_from_symbol"]

        for method in required_methods:
            assert hasattr(IStockForecaster, method)

    def test_forecast_method_signature(self):
        """Test forecast method signature."""
        method = IStockForecaster.forecast
        sig = inspect.signature(method)

        param_names = list(sig.parameters.keys())
        assert "self" in param_names
        assert "data" in param_names


class TestIBacktestRunner:
    """Test IBacktestRunner interface definition."""

    def test_is_abstract_base_class(self):
        """Test that IBacktestRunner is an abstract base class."""
        assert issubclass(IBacktestRunner, ABC)

    def test_has_required_methods(self):
        """Test that IBacktestRunner has all required abstract methods."""
        required_methods = ["run", "run_from_symbol"]

        for method in required_methods:
            assert hasattr(IBacktestRunner, method)

    def test_run_method_signature(self):
        """Test run method signature."""
        method = IBacktestRunner.run
        sig = inspect.signature(method)

        param_names = list(sig.parameters.keys())
        assert "self" in param_names
        assert "data" in param_names
        assert "strategy" in param_names


class TestIDomainFactory:
    """Test IDomainFactory interface definition."""

    def test_is_abstract_base_class(self):
        """Test that IDomainFactory is an abstract base class."""
        assert issubclass(IDomainFactory, ABC)

    def test_has_required_methods(self):
        """Test that IDomainFactory has all required abstract methods."""
        required_methods = ["create_portfolio", "create_asset"]

        for method in required_methods:
            assert hasattr(IDomainFactory, method)

    def test_create_portfolio_signature(self):
        """Test create_portfolio method signature."""
        method = IDomainFactory.create_portfolio
        sig = inspect.signature(method)

        param_names = list(sig.parameters.keys())
        assert "self" in param_names


class TestILoggingManager:
    """Test ILoggingManager interface definition."""

    def test_is_abstract_base_class(self):
        """Test that ILoggingManager is an abstract base class."""
        assert issubclass(ILoggingManager, ABC)

    def test_has_required_methods(self):
        """Test that ILoggingManager has all required abstract methods."""
        required_methods = ["setup", "debug", "info", "warning", "error", "critical"]

        for method in required_methods:
            assert hasattr(ILoggingManager, method)

    def test_logging_methods_signatures(self):
        """Test logging method signatures."""
        for method_name in ["debug", "info", "warning", "error", "critical"]:
            method = getattr(ILoggingManager, method_name)
            sig = inspect.signature(method)

            param_names = list(sig.parameters.keys())
            assert "self" in param_names
            assert "message" in param_names


class TestITechnicalIndicators:
    """Test ITechnicalIndicators interface definition."""

    def test_is_abstract_base_class(self):
        """Test that ITechnicalIndicators is an abstract base class."""
        assert issubclass(ITechnicalIndicators, ABC)

    def test_has_required_methods(self):
        """Test that ITechnicalIndicators has all required abstract methods."""
        required_methods = ["sma", "ema", "rsi", "macd"]

        for method in required_methods:
            assert hasattr(ITechnicalIndicators, method)

    def test_cannot_instantiate_directly(self):
        """Test that ITechnicalIndicators cannot be instantiated directly."""
        try:
            ITechnicalIndicators()
            raise AssertionError("Should not be able to instantiate abstract class")
        except TypeError:
            pass


class TestInterfaceIntegration:
    """Test interface integration and consistency."""

    def test_all_interfaces_importable(self):
        """Test that all interfaces can be imported."""
        from stockula.interfaces import (
            IBacktestRunner,
            IDatabaseManager,
            IDataFetcher,
            IDomainFactory,
            ILoggingManager,
            IStockForecaster,
            ITechnicalIndicators,
        )

        # Just checking they exist and are classes
        for interface in [
            IDataFetcher,
            IDatabaseManager,
            IStockForecaster,
            IBacktestRunner,
            IDomainFactory,
            ILoggingManager,
            ITechnicalIndicators,
        ]:
            assert inspect.isclass(interface)
            assert issubclass(interface, ABC)

    def test_interface_method_consistency(self):
        """Test that interface methods have consistent patterns."""
        interfaces = [
            IDataFetcher,
            IDatabaseManager,
            IStockForecaster,
            IBacktestRunner,
            IDomainFactory,
            ILoggingManager,
            ITechnicalIndicators,
        ]

        for interface in interfaces:
            # All interfaces should have at least one abstract method
            assert len(interface.__abstractmethods__) > 0

            # All abstract methods should be callable
            for method_name in interface.__abstractmethods__:
                method = getattr(interface, method_name)
                assert callable(method)

    def test_interface_inheritance_structure(self):
        """Test interface inheritance structure."""
        # All our interfaces should inherit from ABC
        interfaces = [
            IDataFetcher,
            IDatabaseManager,
            IStockForecaster,
            IBacktestRunner,
            IDomainFactory,
            ILoggingManager,
            ITechnicalIndicators,
        ]

        for interface in interfaces:
            assert ABC in interface.__mro__
            assert interface.__module__ == "stockula.interfaces"


class TestInterfaceUsage:
    """Test how interfaces would be used in practice."""

    def test_interface_implementation_pattern(self):
        """Test that interfaces can be properly implemented."""

        # Create a mock implementation
        class MockDataFetcher(IDataFetcher):
            def get_stock_data(self, symbol, start=None, end=None, interval="1d", force_refresh=False):
                return pd.DataFrame()

            def get_current_prices(self, symbols, show_progress=True):
                return {}

            def get_info(self, symbol, force_refresh=False):
                return {}

            def get_treasury_rates(self, start_date=None, end_date=None, duration="3_month"):
                return pd.Series()

        # Should be able to instantiate the implementation
        mock_fetcher = MockDataFetcher()
        assert isinstance(mock_fetcher, IDataFetcher)

    def test_interface_type_checking(self):
        """Test interface type checking functionality."""
        # Mock implementation for testing
        mock_fetcher = Mock(spec=IDataFetcher)

        # Should pass isinstance check
        assert isinstance(mock_fetcher, type(mock_fetcher))

        # Should have the interface methods
        for method in IDataFetcher.__abstractmethods__:
            assert hasattr(mock_fetcher, method)



"""Centralized logging management for Stockula."""

import logging
import sys
from logging.handlers import RotatingFileHandler

from ..config import StockulaConfig


class LoggingManager:
    """Manages all logging operations for the Stockula application."""

    def __init__(self, name: str = "stockula"):
        """Initialize the LoggingManager.

        Args:
            name: Logger name (default: "stockula")
        """
        self.name = name
        self.logger = logging.getLogger(name)
        self.handlers: list[logging.Handler] = []
        self._is_configured = False

    def setup(self, config: StockulaConfig) -> None:
        """Configure logging based on configuration.

        Args:
            config: Stockula configuration object
        """
        # Prevent duplicate setup
        if self._is_configured:
            return

        # Clear any existing handlers to avoid duplicates
        root_logger = logging.getLogger()
        for handler in root_logger.handlers[:]:
            root_logger.removeHandler(handler)

        # Clear logger's own handlers
        for handler in self.logger.handlers[:]:
            self.logger.removeHandler(handler)

        # List to store all handlers
        self.handlers = []

        # Create console handler
        console_handler = logging.StreamHandler(sys.stdout)

        # Create formatters
        if config.logging.enabled:
            # Detailed format when logging is enabled
            detailed_formatter = logging.Formatter(
                fmt="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
                datefmt="%Y-%m-%d %H:%M:%S",
            )
            simple_formatter = logging.Formatter(fmt="%(asctime)s - %(levelname)s - %(message)s", datefmt="%H:%M:%S")
            log_level = getattr(logging, config.logging.level.upper(), logging.INFO)

            # Use simple formatter for console, detailed for file
            console_handler.setFormatter(simple_formatter)

            # Add file handler if requested
            if config.logging.log_to_file:
                file_handler = RotatingFileHandler(
                    filename=config.logging.log_file,
                    maxBytes=config.logging.max_log_size,
                    backupCount=config.logging.backup_count,
                    encoding="utf-8",
                )
                file_handler.setFormatter(detailed_formatter)
                file_handler.setLevel(log_level)
                self.handlers.append(file_handler)
        else:
            # Simple format when logging is disabled (only warnings/errors)
            formatter = logging.Formatter(fmt="%(levelname)s: %(message)s")
            console_handler.setFormatter(formatter)
            log_level = logging.WARNING

        # Set level on console handler and add to handlers
        console_handler.setLevel(log_level)
        self.handlers.append(console_handler)

        # Configure root logger
        root_logger.setLevel(log_level)
        for handler in self.handlers:
            root_logger.addHandler(handler)

        # Configure stockula loggers with proper hierarchy
        stockula_logger = logging.getLogger("stockula")
        stockula_logger.setLevel(log_level)
        stockula_logger.propagate = True  # Propagate to root logger

        # Store logger reference
        self.logger = stockula_logger

        # Log startup message if enabled
        if config.logging.enabled:
            self.info(f"Logging initialized - Level: {config.logging.level}")
            if config.logging.log_to_file:
                self.info(f"Logging to file: {config.logging.log_file}")

        # Reduce noise from third-party libraries
        third_party_level = (
            logging.CRITICAL
            if not config.logging.enabled
            else (logging.WARNING if log_level != logging.DEBUG else logging.INFO)
        )

        for lib_name in [
            "yfinance",
            "urllib3",
            "requests",
            "apscheduler",
            "peewee",
            "backtesting",
            "cmdstanpy",
            "prophet",
            "prophet.plot",
            "matplotlib",
            "matplotlib.font_manager",
        ]:
            logging.getLogger(lib_name).setLevel(third_party_level)

        self._is_configured = True

    def debug(self, message: str) -> None:
        """Log a debug message.

        Args:
            message: The message to log
        """
        self.logger.debug(message)

    def info(self, message: str) -> None:
        """Log an info message.

        Args:
            message: The message to log
        """
        self.logger.info(message)

    def warning(self, message: str) -> None:
        """Log a warning message.

        Args:
            message: The message to log
        """
        self.logger.warning(message)

    def error(self, message: str, exc_info: bool = False) -> None:
        """Log an error message.

        Args:
            message: The message to log
            exc_info: Whether to include exception info
        """
        self.logger.error(message, exc_info=exc_info)

    def critical(self, message: str, exc_info: bool = False) -> None:
        """Log a critical message.

        Args:
            message: The message to log
            exc_info: Whether to include exception info
        """
        self.logger.critical(message, exc_info=exc_info)

    def exception(self, message: str) -> None:
        """Log an exception with traceback.

        Args:
            message: The message to log
        """
        self.logger.exception(message)

    def isEnabledFor(self, level: int) -> bool:
        """Check if logging is enabled for given level.

        Args:
            level: The logging level to check

        Returns:
            True if logging is enabled for the level
        """
        return self.logger.isEnabledFor(level)

    def get_logger(self) -> logging.Logger:
        """Get the underlying logger instance.

        Returns:
            The logger instance
        """
        return self.logger

    def add_handler(self, handler: logging.Handler) -> None:
        """Add a custom handler to the logger.

        Args:
            handler: The handler to add
        """
        self.handlers.append(handler)
        self.logger.addHandler(handler)

    def remove_handler(self, handler: logging.Handler) -> None:
        """Remove a handler from the logger.

        Args:
            handler: The handler to remove
        """
        if handler in self.handlers:
            self.handlers.remove(handler)
            self.logger.removeHandler(handler)

    def set_level(self, level: str) -> None:
        """Set the logging level.

        Args:
            level: The logging level (DEBUG, INFO, WARNING, ERROR, CRITICAL)
        """
        log_level = getattr(logging, level.upper(), logging.INFO)
        self.logger.setLevel(log_level)
        for handler in self.handlers:
            handler.setLevel(log_level)

    def set_module_level(self, module_name: str, level: str) -> None:
        """Set the logging level for a specific module.

        Args:
            module_name: The module name (e.g., 'cmdstanpy', 'prophet')
            level: The logging level (DEBUG, INFO, WARNING, ERROR, CRITICAL)
        """
        log_level = getattr(logging, level.upper(), logging.WARNING)
        logging.getLogger(module_name).setLevel(log_level)

    def close(self) -> None:
        """Close all handlers and clean up."""
        for handler in self.handlers:
            handler.close()
            self.logger.removeHandler(handler)
        self.handlers.clear()
        self._is_configured = False



"""Forecasting backends for time series prediction."""

from .base import ForecastBackend, ForecastResult
from .simple import SimpleForecastBackend

# Try to import AutoGluon (real dependency), else fall back to simple backend
try:
    import importlib

    importlib.import_module("autogluon.timeseries")
    from .autogluon import AutoGluonBackend

    AUTOGLUON_AVAILABLE = True
except Exception:
    AutoGluonBackend = SimpleForecastBackend  # type: ignore[assignment,misc] # Use simple backend as fallback
    AUTOGLUON_AVAILABLE = False

# Try to import Chronos backend (zero-shot forecasting)
try:
    from .chronos import ChronosBackend

    CHRONOS_AVAILABLE = True
except Exception:  # ImportError or runtime import issues
    ChronosBackend = SimpleForecastBackend  # type: ignore[assignment,misc] # Fallback to simple backend
    CHRONOS_AVAILABLE = False

__all__ = [
    "ForecastBackend",
    "ForecastResult",
    "AutoGluonBackend",
    "SimpleForecastBackend",
    "ChronosBackend",
    "AUTOGLUON_AVAILABLE",
    "CHRONOS_AVAILABLE",
]



"""Strategy Repository - Repository for managing trading strategies with database support."""

from copy import deepcopy
from typing import TYPE_CHECKING, Any, cast

from sqlalchemy.orm import Session

from ..backtesting.strategies import (
    BaseStrategy,
    DoubleEMACrossStrategy,
    FRAMAStrategy,
    KAMAStrategy,
    KaufmanEfficiencyStrategy,
    MACDStrategy,
    RSIStrategy,
    SMACrossStrategy,
    TRIMACrossStrategy,
    TripleEMACrossStrategy,
    VAMAStrategy,
    VIDYAStrategy,
)
from ..database.models import Strategy, StrategyPreset
from .repository import Repository

if TYPE_CHECKING:
    from ..database.manager import DatabaseManager


class StrategyRepository(Repository[type[BaseStrategy]]):
    """Repository for trading strategies with database persistence.

    This class manages all available trading strategies and provides methods
    for strategy discovery, validation, and configuration. It can operate
    in both in-memory mode and database-backed mode.
    """

    # Class property: mappings of strategy names to classes
    DEFAULT_MAPPINGS = {
        "smacross": SMACrossStrategy,
        "rsi": RSIStrategy,
        "macd": MACDStrategy,
        "double_ema_cross": DoubleEMACrossStrategy,
        "triple_ema_cross": TripleEMACrossStrategy,
        "trima_cross": TRIMACrossStrategy,
        "kama": KAMAStrategy,
        "frama": FRAMAStrategy,
        "vama": VAMAStrategy,
        "vidya": VIDYAStrategy,
        "kaufman_efficiency": KaufmanEfficiencyStrategy,
    }

    # Default strategy groups
    DEFAULT_GROUPS = {
        "basic": ["smacross", "rsi"],
        "momentum": ["rsi", "macd", "double_ema_cross"],
        "trend": ["smacross", "triple_ema_cross", "trima_cross"],
        "advanced": ["kama", "frama", "vama", "vidya"],
        "comprehensive": [
            "smacross",
            "rsi",
            "macd",
            "double_ema_cross",
            "triple_ema_cross",
            "trima_cross",
            "kama",
            "frama",
            "vama",
            "vidya",
            "kaufman_efficiency",
        ],
    }

    # Default parameter presets
    DEFAULT_PRESETS = {
        "smacross": {"fast_period": 10, "slow_period": 20},
        "rsi": {"period": 14, "oversold_threshold": 30.0, "overbought_threshold": 70.0},
        "macd": {"fast_period": 12, "slow_period": 26, "signal_period": 9},
        "double_ema_cross": {
            "fast_period": 34,
            "slow_period": 55,
            "momentum_atr_multiple": 1.25,
            "speculative_atr_multiple": 1.0,
            "atr_period": 14,
        },
        "triple_ema_cross": {"fast": 5, "medium": 10, "slow": 20, "fast_period": 9, "slow_period": 21},
        "trima_cross": {"fast_period": 14, "slow_period": 28, "atr_multiple": 1.2},
        "kama": {"period": 14, "fast_sc": 2, "slow_sc": 30, "er_period": 10},
        "frama": {"period": 14, "frama_period": 16},
        "vama": {"period": 8, "vama_period": 8, "slow_vama_period": 21},
        "vidya": {"period": 14, "alpha": 0.2, "cmo_period": 9, "smoothing_period": 12},
        "kaufman_efficiency": {
            "period": 10,
            "fast_sc": 2,
            "slow_sc": 30,
            "er_upper_threshold": 0.5,
            "er_lower_threshold": 0.3,
        },
    }

    def __init__(self, db_manager: "DatabaseManager | None" = None):
        """Initialize the strategy repository.

        Args:
            db_manager: Optional database manager for persistence
        """
        super().__init__()
        self.db_manager = db_manager

        # Initialize groups and values (deep copies to prevent mutation)
        self.groups = deepcopy(self.DEFAULT_GROUPS)
        self._preset_values = deepcopy(self.DEFAULT_PRESETS)

        # Initialize with default strategies
        self._items = self.DEFAULT_MAPPINGS.copy()

        # Load from database if available
        if self.db_manager:
            self._load_from_database()

    def _load_from_database(self) -> None:
        """Load strategies and presets from the database."""
        if not self.db_manager:
            return

        with self.db_manager.get_session() as session:
            # Load active strategies
            # Use type: ignore for SQLAlchemy filter
            strategies = session.query(Strategy).filter(Strategy.is_active == True).all()  # type: ignore[arg-type] # noqa: E712
            for strategy in strategies:
                # For now, we only load presets - actual strategy classes
                # would need to be dynamically imported from module_path
                self._load_strategy_presets(session, strategy)

    def _load_strategy_presets(self, session: Session, strategy: Strategy) -> None:
        """Load presets for a specific strategy from the database."""
        presets = session.query(StrategyPreset).filter(StrategyPreset.strategy_id == strategy.id).all()  # type: ignore[arg-type]

        for preset in presets:
            if preset.is_default:
                self._preset_values[strategy.name] = cast(dict[str, Any], preset.parameters)

    def sync_to_database(self) -> None:
        """Sync current strategies and presets to the database."""
        if not self.db_manager:
            return

        from sqlalchemy.exc import IntegrityError, OperationalError

        try:
            with self.db_manager.get_session() as session:
                # Sync strategies
                for name, strategy_class in self._items.items():
                    existing = session.query(Strategy).filter(Strategy.name == name).first()  # type: ignore[arg-type]
                    if not existing:
                        strategy = Strategy(
                            name=name,
                            class_name=strategy_class.__name__,
                            module_path=strategy_class.__module__,
                            description=strategy_class.__doc__.split("\n")[0] if strategy_class.__doc__ else None,
                            category=self._get_strategy_category(name),
                        )
                        session.add(strategy)
                        session.flush()  # Get the ID

                        # Add default preset
                        if name in self._preset_values:
                            preset = StrategyPreset(
                                strategy_id=strategy.id,
                                name="default",
                                is_default=True,
                            )
                            preset.set_parameters(cast(dict[Any, Any], self._preset_values[name]))
                            session.add(preset)

                session.commit()
        except (IntegrityError, OperationalError):
            # Handle cases where:
            # - Strategy already exists (IntegrityError on unique constraint)
            # - Tables don't exist yet (OperationalError)
            # - Concurrent access from multiple processes/threads
            session.rollback()

    def _get_strategy_category(self, strategy_name: str) -> str | None:
        """Determine the category of a strategy based on groups."""
        for group_name, strategies in self.groups.items():
            if strategy_name in strategies and group_name != "comprehensive":
                return group_name
        return None

    @property
    def presets(self) -> dict[str, dict[str, Any]]:
        """Get immutable default presets.

        Returns:
            Dictionary of default presets (read-only deep copy)
        """
        return cast(dict[str, dict[str, Any]], deepcopy(self.DEFAULT_PRESETS))

    @property
    def values(self) -> dict[str, dict[str, Any]]:  # type: ignore[override]
        """Get mutable preset values.

        Returns:
            Dictionary of current preset values (can be modified)
        """
        return cast(dict[str, dict[str, Any]], self._preset_values)

    def add(self, name: str, strategy_class: type[BaseStrategy]) -> None:
        """Add a strategy to the repository.

        Args:
            name: Strategy name (will be converted to lowercase)
            strategy_class: The strategy class
        """
        lowercase_name = name.lower()
        self._items[lowercase_name] = strategy_class

    def get(self, key: str, default=None):
        """Get a strategy by name (case-insensitive).

        Args:
            key: Strategy name in any case
            default: Default value if not found

        Returns:
            Strategy class or default
        """
        return self._items.get(key.lower(), default)

    def remove(self, key: str) -> None:
        """Remove a strategy from the repository.

        Args:
            key: Strategy name to remove

        Raises:
            KeyError: If strategy not found
        """
        lowercase_key = key.lower()
        if lowercase_key not in self._items:
            raise KeyError(f"Strategy '{key}' not found")
        del self._items[lowercase_key]

    def validate(self, name: str) -> None:
        """Validate that a strategy name exists.

        Args:
            name: Strategy name to validate

        Raises:
            ValueError: If the strategy name is not valid
        """
        lowercase_name = name.lower()
        if lowercase_name not in self._items:
            available = list(self._items.keys())
            raise ValueError(f"Unknown strategy: {name}. Available strategies: {available}")

    def get_strategy_class(self, strategy_name: str) -> type[BaseStrategy] | None:
        """Get a strategy class by name.

        Args:
            strategy_name: Strategy name (case-insensitive)

        Returns:
            Strategy class or None if not found
        """
        return cast(type[BaseStrategy] | None, self.get(strategy_name))

    def normalize_strategy_name(self, strategy_name: str) -> str:
        """Normalize strategy name to lowercase format.

        Args:
            strategy_name: Strategy name in any format

        Returns:
            Normalized (lowercase) strategy name
        """
        return strategy_name.lower()

    def get_available_strategy_names(self) -> list[str]:
        """Get list of all available strategy names.

        Returns:
            List of all strategy names
        """
        return list(self._items.keys())

    def is_valid_strategy(self, strategy_name: str) -> bool:
        """Check if a strategy name is valid.

        Args:
            strategy_name: Strategy name to validate

        Returns:
            True if valid, False otherwise
        """
        return strategy_name.lower() in self._items

    def get_strategy_groups(self) -> dict[str, list[str]]:
        """Get all strategy groups.

        Returns:
            Dictionary of group names to strategy lists
        """
        return self.groups.copy()

    def is_valid_strategy_group(self, group_name: str) -> bool:
        """Check if a strategy group is valid.

        Args:
            group_name: Group name to validate

        Returns:
            True if valid, False otherwise
        """
        return group_name in self.groups

    def get_strategies_in_group(self, group_name: str) -> list[str]:
        """Get strategies in a specific group.

        Args:
            group_name: Name of the group

        Returns:
            List of strategy names in the group

        Raises:
            ValueError: If group name is invalid
        """
        if not self.is_valid_strategy_group(group_name):
            available_groups = list(self.groups.keys())
            raise ValueError(f"Unknown strategy group: {group_name}. Available: {available_groups}")

        return self.groups[group_name].copy()

    def get_strategy_presets(self) -> dict[str, dict[str, Any]]:
        """Get all strategy parameter presets.

        Returns:
            Dictionary of strategy names to parameter presets
        """
        return cast(dict[str, dict[str, Any]], self._preset_values.copy())

    def get_strategy_preset(self, strategy_name: str) -> dict[str, Any]:
        """Get parameter preset for a specific strategy.

        Args:
            strategy_name: Strategy name

        Returns:
            Dictionary of default parameters
        """
        normalized_name = self.normalize_strategy_name(strategy_name)
        preset = self._preset_values.get(normalized_name, {})
        return cast(dict[str, Any], preset.copy() if isinstance(preset, dict) else {})

    def validate_strategies(self, strategy_names: list[str]) -> tuple[list[str], list[str]]:
        """Validate a list of strategy names.

        Args:
            strategy_names: List of strategy names to validate

        Returns:
            Tuple of (valid_strategies, invalid_strategies) - valid names are normalized to lowercase
        """
        valid = []
        invalid = []

        for name in strategy_names:
            lowercase_name = name.lower()
            if lowercase_name in self._items:
                valid.append(lowercase_name)
            else:
                invalid.append(name)

        return valid, invalid

    def add_strategy_group(self, group_name: str, strategies: list[str]) -> None:
        """Add or update a strategy group.

        Args:
            group_name: Name of the group
            strategies: List of strategy names

        Raises:
            ValueError: If any strategy is invalid
        """
        # Validate all strategies first
        for strategy in strategies:
            self.validate(strategy)

        # Store normalized (lowercase) names
        self.groups[group_name] = [s.lower() for s in strategies]

    def update_strategy_preset(self, strategy_name: str, parameters: dict[str, Any]) -> None:
        """Update parameters for a strategy preset.

        Args:
            strategy_name: Strategy name
            parameters: Parameters to update

        Raises:
            ValueError: If strategy is invalid
        """
        self.validate(strategy_name)

        lowercase_name = strategy_name.lower()
        if lowercase_name not in self._preset_values:
            self._preset_values[lowercase_name] = {}

        # Cast to dict for type safety
        preset = cast(dict[str, Any], self._preset_values[lowercase_name])
        preset.update(parameters)

        # Sync to database if available
        if self.db_manager:
            self._save_preset_to_database(lowercase_name, cast(dict[str, Any], self._preset_values[lowercase_name]))

    def _save_preset_to_database(self, strategy_name: str, parameters: dict[str, Any]) -> None:
        """Save a preset to the database."""
        if not self.db_manager:
            return

        with self.db_manager.get_session() as session:
            strategy = session.query(Strategy).filter(Strategy.name == strategy_name).first()  # type: ignore[arg-type]
            if strategy:
                preset = (
                    session.query(StrategyPreset)
                    .filter(StrategyPreset.strategy_id == strategy.id, StrategyPreset.name == "default")  # type: ignore[arg-type]
                    .first()
                )

                if preset:
                    preset.set_parameters(parameters)
                else:
                    preset = StrategyPreset(
                        strategy_id=strategy.id,
                        name="default",
                        is_default=True,
                    )
                    preset.set_parameters(parameters)
                    session.add(preset)

                session.commit()


# Create a singleton instance (without database for backward compatibility)
strategy_repository = StrategyRepository()



"""SQLModel models for the Stockula database.

This module defines database models using SQLModel, which combines
SQLAlchemy ORM with Pydantic validation.
"""

import json
from datetime import UTC, datetime
from datetime import date as DateType
from typing import Any, ClassVar, Optional, cast

from sqlalchemy import DateTime as SQLADateTime
from sqlalchemy import func
from sqlmodel import Column, Field, Index, Relationship, SQLModel, Text, UniqueConstraint


class Stock(SQLModel, table=True):  # type: ignore[call-arg]
    """Basic stock metadata."""

    __tablename__ = "stocks"

    symbol: str = Field(primary_key=True, description="Stock ticker symbol")
    name: str | None = Field(default=None, description="Company name")
    sector: str | None = Field(default=None, description="Business sector")
    industry: str | None = Field(default=None, description="Industry classification")
    market_cap: float | None = Field(default=None, description="Market capitalization")
    exchange: str | None = Field(default=None, description="Stock exchange")
    currency: str | None = Field(default=None, description="Trading currency")
    created_at: datetime = Field(
        default_factory=lambda: datetime.now(UTC),
        sa_column=Column(SQLADateTime, server_default=func.current_timestamp()),
        description="Timestamp when the record was created",
    )
    updated_at: datetime = Field(
        default_factory=lambda: datetime.now(UTC),
        sa_column=Column(
            SQLADateTime,
            server_default=func.current_timestamp(),
            onupdate=func.current_timestamp(),
        ),
        description="Timestamp when the record was last updated",
    )

    # Relationships
    price_history: list["PriceHistory"] = Relationship(back_populates="stock", cascade_delete=True)
    dividends: list["Dividend"] = Relationship(back_populates="stock", cascade_delete=True)
    splits: list["Split"] = Relationship(back_populates="stock", cascade_delete=True)
    options_calls: list["OptionsCall"] = Relationship(back_populates="stock", cascade_delete=True)
    options_puts: list["OptionsPut"] = Relationship(back_populates="stock", cascade_delete=True)
    stock_info: Optional["StockInfo"] = Relationship(back_populates="stock", cascade_delete=True)


class PriceHistory(SQLModel, table=True):  # type: ignore[call-arg]
    """Historical OHLCV price data."""

    __tablename__ = "price_history"
    __table_args__ = (
        UniqueConstraint("symbol", "date", "interval", name="uq_price_history"),
        Index("idx_price_history_symbol_date", "symbol", "date"),
    )

    id: int | None = Field(default=None, primary_key=True, description="Primary key")
    symbol: str = Field(foreign_key="stocks.symbol", description="Stock ticker symbol")
    date: DateType = Field(description="Trading date")
    open_price: float | None = Field(default=None, description="Opening price")
    high_price: float | None = Field(default=None, description="Highest price")
    low_price: float | None = Field(default=None, description="Lowest price")
    close_price: float | None = Field(default=None, description="Closing price")
    volume: int | None = Field(default=None, description="Trading volume")
    interval: str = Field(default="1d", description="Data interval (1d, 1h, etc.)")
    created_at: datetime = Field(
        default_factory=lambda: datetime.now(UTC),
        sa_column=Column(SQLADateTime, server_default=func.current_timestamp()),
        description="Timestamp when the record was created",
    )
    updated_at: datetime = Field(
        default_factory=lambda: datetime.now(UTC),
        sa_column=Column(
            SQLADateTime,
            server_default=func.current_timestamp(),
            onupdate=func.current_timestamp(),
        ),
        description="Timestamp when the record was last updated",
    )

    # Relationships
    stock: Stock = Relationship(back_populates="price_history")


class Dividend(SQLModel, table=True):  # type: ignore[call-arg]
    """Dividend payment history."""

    __tablename__ = "dividends"
    __table_args__ = (
        UniqueConstraint("symbol", "date", name="uq_dividends"),
        Index("idx_dividends_symbol_date", "symbol", "date"),
    )

    id: int | None = Field(default=None, primary_key=True, description="Primary key")
    symbol: str = Field(foreign_key="stocks.symbol", description="Stock ticker symbol")
    date: DateType = Field(description="Dividend payment date")
    amount: float = Field(description="Dividend amount per share")
    created_at: datetime = Field(
        default_factory=lambda: datetime.now(UTC),
        sa_column=Column(SQLADateTime, server_default=func.current_timestamp()),
        description="Timestamp when the record was created",
    )
    updated_at: datetime = Field(
        default_factory=lambda: datetime.now(UTC),
        sa_column=Column(
            SQLADateTime,
            server_default=func.current_timestamp(),
            onupdate=func.current_timestamp(),
        ),
        description="Timestamp when the record was last updated",
    )

    # Relationships
    stock: Stock = Relationship(back_populates="dividends")


class Split(SQLModel, table=True):  # type: ignore[call-arg]
    """Stock split history."""

    __tablename__ = "splits"
    __table_args__ = (
        UniqueConstraint("symbol", "date", name="uq_splits"),
        Index("idx_splits_symbol_date", "symbol", "date"),
    )

    id: int | None = Field(default=None, primary_key=True, description="Primary key")
    symbol: str = Field(foreign_key="stocks.symbol", description="Stock ticker symbol")
    date: DateType = Field(description="Split date")
    ratio: float = Field(description="Split ratio (e.g., 2.0 for 2:1 split)")
    created_at: datetime = Field(
        default_factory=lambda: datetime.now(UTC),
        sa_column=Column(SQLADateTime, server_default=func.current_timestamp()),
        description="Timestamp when the record was created",
    )
    updated_at: datetime = Field(
        default_factory=lambda: datetime.now(UTC),
        sa_column=Column(
            SQLADateTime,
            server_default=func.current_timestamp(),
            onupdate=func.current_timestamp(),
        ),
        description="Timestamp when the record was last updated",
    )

    # Relationships
    stock: Stock = Relationship(back_populates="splits")


class OptionsCall(SQLModel, table=True):  # type: ignore[call-arg]
    """Call options data."""

    __tablename__ = "options_calls"
    __table_args__ = (
        UniqueConstraint(
            "symbol",
            "expiration_date",
            "strike",
            "contract_symbol",
            name="uq_options_calls",
        ),
        Index("idx_options_calls_symbol_exp", "symbol", "expiration_date"),
    )

    id: int | None = Field(default=None, primary_key=True, description="Primary key")
    symbol: str = Field(foreign_key="stocks.symbol", description="Stock ticker symbol")
    expiration_date: DateType = Field(description="Option expiration date")
    strike: float = Field(description="Strike price")
    last_price: float | None = Field(default=None, description="Last traded price")
    bid: float | None = Field(default=None, description="Bid price")
    ask: float | None = Field(default=None, description="Ask price")
    volume: int | None = Field(default=None, description="Trading volume")
    open_interest: int | None = Field(default=None, description="Open interest")
    implied_volatility: float | None = Field(default=None, description="Implied volatility")
    in_the_money: bool | None = Field(default=None, description="Whether option is in the money")
    contract_symbol: str | None = Field(default=None, description="Option contract symbol")
    created_at: datetime = Field(
        default_factory=lambda: datetime.now(UTC),
        sa_column=Column(SQLADateTime, server_default=func.current_timestamp()),
        description="Timestamp when the record was created",
    )
    updated_at: datetime = Field(
        default_factory=lambda: datetime.now(UTC),
        sa_column=Column(
            SQLADateTime,
            server_default=func.current_timestamp(),
            onupdate=func.current_timestamp(),
        ),
        description="Timestamp when the record was last updated",
    )

    # Relationships
    stock: Stock = Relationship(back_populates="options_calls")


class OptionsPut(SQLModel, table=True):  # type: ignore[call-arg]
    """Put options data."""

    __tablename__ = "options_puts"
    __table_args__ = (
        UniqueConstraint(
            "symbol",
            "expiration_date",
            "strike",
            "contract_symbol",
            name="uq_options_puts",
        ),
        Index("idx_options_puts_symbol_exp", "symbol", "expiration_date"),
    )

    id: int | None = Field(default=None, primary_key=True, description="Primary key")
    symbol: str = Field(foreign_key="stocks.symbol", description="Stock ticker symbol")
    expiration_date: DateType = Field(description="Option expiration date")
    strike: float = Field(description="Strike price")
    last_price: float | None = Field(default=None, description="Last traded price")
    bid: float | None = Field(default=None, description="Bid price")
    ask: float | None = Field(default=None, description="Ask price")
    volume: int | None = Field(default=None, description="Trading volume")
    open_interest: int | None = Field(default=None, description="Open interest")
    implied_volatility: float | None = Field(default=None, description="Implied volatility")
    in_the_money: bool | None = Field(default=None, description="Whether option is in the money")
    contract_symbol: str | None = Field(default=None, description="Option contract symbol")
    created_at: datetime = Field(
        default_factory=lambda: datetime.now(UTC),
        sa_column=Column(SQLADateTime, server_default=func.current_timestamp()),
        description="Timestamp when the record was created",
    )
    updated_at: datetime = Field(
        default_factory=lambda: datetime.now(UTC),
        sa_column=Column(
            SQLADateTime,
            server_default=func.current_timestamp(),
            onupdate=func.current_timestamp(),
        ),
        description="Timestamp when the record was last updated",
    )

    # Relationships
    stock: Stock = Relationship(back_populates="options_puts")


class StockInfo(SQLModel, table=True):  # type: ignore[call-arg]
    """Raw yfinance info data stored as JSON."""

    __tablename__ = "stock_info"

    symbol: str = Field(foreign_key="stocks.symbol", primary_key=True, description="Stock ticker symbol")
    info_json: str = Field(
        sa_column=Column(Text, nullable=False),
        description="JSON-encoded stock information from yfinance",
    )
    created_at: datetime = Field(
        default_factory=lambda: datetime.now(UTC),
        sa_column=Column(SQLADateTime, server_default=func.current_timestamp()),
        description="Timestamp when the record was created",
    )
    updated_at: datetime = Field(
        default_factory=lambda: datetime.now(UTC),
        sa_column=Column(
            SQLADateTime,
            server_default=func.current_timestamp(),
            onupdate=func.current_timestamp(),
        ),
        description="Timestamp when the record was last updated",
    )

    # Relationships
    stock: Stock = Relationship(back_populates="stock_info")

    @property
    def info_dict(self) -> dict[str, Any]:
        """Parse JSON info to dictionary."""
        return cast(dict[str, Any], json.loads(self.info_json))

    def set_info(self, info: dict) -> None:
        """Set info from dictionary."""
        self.info_json = json.dumps(info, default=str)


class Strategy(SQLModel, table=True):  # type: ignore[call-arg]
    """Model for trading strategies."""

    __tablename__ = "strategies"

    id: int | None = Field(default=None, primary_key=True)
    name: str = Field(index=True, unique=True)
    class_name: str = Field()  # Full class name (e.g., 'SMACrossStrategy')
    module_path: str = Field()  # Module path for importing
    description: str | None = Field(default=None)
    category: str | None = Field(default=None)  # e.g., 'momentum', 'trend'
    is_active: bool = Field(default=True)
    created_at: datetime = Field(default_factory=lambda: datetime.now(UTC))
    updated_at: datetime = Field(default_factory=lambda: datetime.now(UTC))

    # Relationships
    presets: list["StrategyPreset"] = Relationship(back_populates="strategy")


class StrategyPreset(SQLModel, table=True):  # type: ignore[call-arg]
    """Model for strategy parameter presets."""

    __tablename__ = "strategy_presets"

    id: int | None = Field(default=None, primary_key=True)
    strategy_id: int = Field(foreign_key="strategies.id", index=True)
    name: str = Field(index=True)  # e.g., 'default', 'conservative', 'aggressive'
    parameters_json: str = Field()  # JSON string of parameters
    description: str | None = Field(default=None)
    is_default: bool = Field(default=False)
    created_at: datetime = Field(default_factory=lambda: datetime.now(UTC))
    updated_at: datetime = Field(default_factory=lambda: datetime.now(UTC))

    # Relationships
    strategy: Strategy = Relationship(back_populates="presets")

    # Unique constraint on strategy_id + name
    __table_args__ = (UniqueConstraint("strategy_id", "name", name="unique_strategy_preset"),)

    @property
    def parameters(self) -> dict[str, Any]:
        """Parse JSON parameters to dictionary."""
        return cast(dict[str, Any], json.loads(self.parameters_json))

    def set_parameters(self, params: dict) -> None:
        """Set parameters from dictionary."""
        self.parameters_json = json.dumps(params, default=str)


class AutoTSModel(SQLModel, table=True):  # type: ignore[call-arg]
    """AutoTS model definitions and metadata.

    This model stores valid AutoTS models and their properties.
    Validation is performed before saving to ensure only valid models are stored.
    """

    __tablename__ = "autots_models"

    # Class-level registry of valid models (loaded from models.json)
    _valid_models: ClassVar[set[str] | None] = None
    _models_data: ClassVar[dict[str, dict] | None] = None

    id: int | None = Field(default=None, primary_key=True)
    name: str = Field(index=True, unique=True, description="Model name (e.g., ARIMA, ETS)")
    categories: str = Field(default="[]", description="JSON array of categories (univariate, multivariate, etc.)")
    is_slow: bool = Field(default=False, description="Whether model is computationally expensive")
    is_gpu_enabled: bool = Field(default=False, description="Whether model can utilize GPU")
    requires_regressor: bool = Field(default=False, description="Whether model supports external regressors")
    min_data_points: int = Field(default=10, description="Minimum data points required")
    description: str | None = Field(default=None, description="Model description")
    created_at: datetime = Field(default_factory=datetime.utcnow)
    updated_at: datetime = Field(default_factory=datetime.utcnow)

    @classmethod
    def load_valid_models(cls, force_reload: bool = False) -> None:
        """Load valid models from AutoTS library.

        Args:
            force_reload: Force reload even if already loaded
        """
        if cls._valid_models is not None and not force_reload:
            return

        try:
            # Get the authoritative model list from AutoTS itself
            from autots.models.model_list import model_lists

            # The model_lists is a dictionary of presets
            # The 'all' preset contains all available models
            all_models = set(model_lists.get("all", []))

            cls._valid_models = all_models
            cls._models_data = model_lists  # Store the full model_lists for reference

        except ImportError:
            # Fallback if AutoTS is not available (shouldn't happen in normal operation)
            # Use a minimal set of known models
            cls._valid_models = {
                "ARIMA",
                "ETS",
                "FBProphet",
                "GluonTS",
                "VAR",
                "VECM",
                "Theta",
                "UnivariateMotif",
                "MultivariateMotif",
                "LastValueNaive",
                "ConstantNaive",
                "AverageValueNaive",
                "SeasonalNaive",
                "GLM",
                "GLS",
                "RollingRegression",
                "UnobservedComponents",
                "DynamicFactor",
                "WindowRegression",
                "DatepartRegression",
                "UnivariateRegression",
                "NVAR",
                "MultivariateRegression",
                "ARDL",
                "NeuralProphet",
            }
            cls._models_data = {}

    @classmethod
    def is_valid_model(cls, name: str) -> bool:
        """Check if a model name is valid.

        Args:
            name: Model name to validate

        Returns:
            True if model is valid, False otherwise
        """
        cls.load_valid_models()
        return name in (cls._valid_models or set())

    @classmethod
    def get_valid_models(cls) -> set[str]:
        """Get set of all valid model names.

        Returns:
            Set of valid model names
        """
        cls.load_valid_models()
        return cls._valid_models.copy() if cls._valid_models else set()

    @classmethod
    def validate_model_list(cls, models: list[str]) -> tuple[bool, list[str]]:
        """Validate a list of model names.

        Args:
            models: List of model names to validate

        Returns:
            Tuple of (all_valid, invalid_models)
        """
        cls.load_valid_models()
        invalid = [m for m in models if not cls.is_valid_model(m)]
        return len(invalid) == 0, invalid

    def validate_model(self) -> None:
        """Validate the model before saving.

        Raises:
            ValueError: If the model name is not a valid AutoTS model
        """
        if not self.is_valid_model(self.name):
            valid_models = self.get_valid_models()
            raise ValueError(
                f"'{self.name}' is not a valid AutoTS model. "
                f"Valid models include: {', '.join(sorted(list(valid_models)[:10]))}..."
            )

    @property
    def category_list(self) -> list[str]:
        """Parse JSON categories to list."""
        return cast(list[str], json.loads(self.categories))

    def set_categories(self, categories: list[str]) -> None:
        """Set categories from list."""
        self.categories = json.dumps(categories)


class AutoTSPreset(SQLModel, table=True):  # type: ignore[call-arg]
    """AutoTS model preset configurations.

    This model stores preset configurations that group models together.
    Validation ensures all models in a preset are valid AutoTS models.
    """

    __tablename__ = "autots_presets"

    # Class-level registry of valid presets
    _valid_presets: ClassVar[set[str] | None] = None

    id: int | None = Field(default=None, primary_key=True)
    name: str = Field(index=True, unique=True, description="Preset name (e.g., fast, superfast)")
    models: str = Field(description="JSON array of model names or dict with weights")
    description: str | None = Field(default=None, description="Preset description")
    use_case: str | None = Field(default=None, description="Recommended use case")
    created_at: datetime = Field(default_factory=datetime.utcnow)
    updated_at: datetime = Field(default_factory=datetime.utcnow)

    @classmethod
    def load_valid_presets(cls, force_reload: bool = False) -> None:
        """Load valid presets from AutoTS library.

        Args:
            force_reload: Force reload even if already loaded
        """
        if cls._valid_presets is not None and not force_reload:
            return

        try:
            # Get presets directly from AutoTS
            from autots.models.model_list import model_lists

            # All keys in model_lists are valid presets
            cls._valid_presets = set(model_lists.keys())
        except ImportError:
            # Fallback to known AutoTS presets
            cls._valid_presets = {
                "fast",
                "superfast",
                "default",
                "parallel",
                "fast_parallel",
                "scalable",
                "probabilistic",
                "multivariate",
                "univariate",
                "best",
                "slow",
                "gpu",
                "regressor",
                "motifs",
                "regressions",
                "all",
                "no_shared",
                "experimental",
            }

    @classmethod
    def is_valid_preset(cls, name: str) -> bool:
        """Check if a preset name is valid.

        Args:
            name: Preset name to validate

        Returns:
            True if preset is valid, False otherwise
        """
        cls.load_valid_presets()
        return name in (cls._valid_presets or set())

    def validate_model(self) -> None:
        """Validate the preset before saving.

        Raises:
            ValueError: If preset contains invalid models
        """
        # Parse the models
        model_list = self.model_list

        # Extract model names depending on format
        if isinstance(model_list, list):
            model_names = model_list
        elif isinstance(model_list, dict):
            model_names = list(model_list.keys())
        else:
            raise ValueError(f"Invalid models format: {type(model_list)}")

        # Validate all models in the preset
        is_valid, invalid_models = AutoTSModel.validate_model_list(model_names)
        if not is_valid:
            raise ValueError(f"Preset '{self.name}' contains invalid models: {', '.join(invalid_models)}")

    @property
    def model_list(self) -> list[str] | dict[str, float]:
        """Parse JSON models to list or dict."""
        data = json.loads(self.models)
        return cast("list[str] | dict[str, float]", data)

    def set_models(self, models: list[str] | dict[str, float]) -> None:
        """Set models from list or dict."""
        self.models = json.dumps(models)



"""Unit tests for AllocatorManager."""

from unittest.mock import MagicMock

import pytest

from stockula.allocation import AllocatorManager
from stockula.config import TickerConfig


@pytest.fixture
def mock_data_fetcher():
    """Create a mock data fetcher."""
    fetcher = MagicMock()
    fetcher.get_current_prices.return_value = {"AAPL": 150.0, "GOOGL": 2800.0, "MSFT": 300.0}
    fetcher.get_stock_data.return_value = MagicMock()
    fetcher.get_stock_info.return_value = {"marketCap": 1000000000}
    return fetcher


@pytest.fixture
def mock_backtest_runner():
    """Create a mock backtest runner."""
    return MagicMock()


@pytest.fixture
def mock_logging_manager():
    """Create a mock logging manager."""
    return MagicMock()


@pytest.fixture
def sample_config():
    """Create a sample configuration."""
    config = MagicMock()
    # Mock portfolio attributes
    config.portfolio = MagicMock()
    config.portfolio.initial_capital = 100000
    config.portfolio.allocation_method = "equal_weight"
    config.portfolio.allow_fractional_shares = True
    config.portfolio.capital_utilization_target = 0.95
    config.portfolio.category_ratios = {"TECH": 0.6, "FINANCE": 0.4}
    config.portfolio.tickers = []
    # Mock data attributes
    config.data = MagicMock()
    config.data.start_date = "2024-01-01"
    config.data.end_date = "2024-12-31"
    # Mock backtest_optimization
    config.backtest_optimization = MagicMock()
    return config


@pytest.fixture
def sample_tickers():
    """Create sample ticker configurations."""
    return [
        TickerConfig(symbol="AAPL", category="TECH", allocation_pct=50),
        TickerConfig(symbol="GOOGL", category="TECH", allocation_pct=30),
        TickerConfig(symbol="MSFT", category="TECH", allocation_pct=20),
    ]


class TestAllocatorManager:
    """Test cases for AllocatorManager."""

    def test_init(self, mock_data_fetcher, mock_backtest_runner, mock_logging_manager):
        """Test AllocatorManager initialization."""
        manager = AllocatorManager(
            data_fetcher=mock_data_fetcher,
            backtest_runner=mock_backtest_runner,
            logging_manager=mock_logging_manager,
        )

        assert manager.data_fetcher == mock_data_fetcher
        assert manager.backtest_runner == mock_backtest_runner
        assert manager.logger == mock_logging_manager
        assert manager.standard_allocator is not None
        assert manager.backtest_allocator is not None
        assert len(manager.allocator_map) == 6

    def test_get_allocator(self, mock_data_fetcher, mock_backtest_runner, mock_logging_manager):
        """Test getting allocator by method."""
        manager = AllocatorManager(
            data_fetcher=mock_data_fetcher,
            backtest_runner=mock_backtest_runner,
            logging_manager=mock_logging_manager,
        )

        # Test standard allocator methods
        assert manager.get_allocator("equal_weight") == manager.standard_allocator
        assert manager.get_allocator("market_cap") == manager.standard_allocator
        assert manager.get_allocator("custom") == manager.standard_allocator
        assert manager.get_allocator("dynamic") == manager.standard_allocator
        assert manager.get_allocator("auto") == manager.standard_allocator

        # Test backtest allocator
        assert manager.get_allocator("backtest_optimized") == manager.backtest_allocator

        # Test unknown method
        with pytest.raises(ValueError, match="Unknown allocation method"):
            manager.get_allocator("unknown_method")

    def test_calculate_quantities(
        self, mock_data_fetcher, mock_backtest_runner, mock_logging_manager, sample_config, sample_tickers
    ):
        """Test calculate_quantities delegates to appropriate allocator."""
        manager = AllocatorManager(
            data_fetcher=mock_data_fetcher,
            backtest_runner=mock_backtest_runner,
            logging_manager=mock_logging_manager,
        )

        # Mock the standard allocator's calculate_quantities method
        expected_result = {"AAPL": 222.22, "GOOGL": 11.90, "MSFT": 111.11}
        manager.standard_allocator.calculate_quantities = MagicMock(return_value=expected_result)

        result = manager.calculate_quantities(sample_config, sample_tickers)

        assert result == expected_result
        manager.standard_allocator.calculate_quantities.assert_called_once_with(sample_config, sample_tickers)

    def test_calculate_equal_weight_quantities(
        self, mock_data_fetcher, mock_backtest_runner, mock_logging_manager, sample_config, sample_tickers
    ):
        """Test equal weight quantity calculation."""
        manager = AllocatorManager(
            data_fetcher=mock_data_fetcher,
            backtest_runner=mock_backtest_runner,
            logging_manager=mock_logging_manager,
        )

        result = manager.calculate_equal_weight_quantities(sample_config, sample_tickers)

        # With $100,000 split equally among 3 stocks
        # AAPL at $150: $33,333.33 / $150 = 222.22 shares
        # GOOGL at $2800: $33,333.33 / $2800 = 11.90 shares
        # MSFT at $300: $33,333.33 / $300 = 111.11 shares
        assert result["AAPL"] == pytest.approx(222.22, rel=0.01)
        assert result["GOOGL"] == pytest.approx(11.90, rel=0.01)
        assert result["MSFT"] == pytest.approx(111.11, rel=0.01)

    def test_calculate_market_cap_quantities(
        self, mock_data_fetcher, mock_backtest_runner, mock_logging_manager, sample_config, sample_tickers
    ):
        """Test market cap weighted quantity calculation."""
        # Set up market caps
        mock_data_fetcher.get_stock_info.side_effect = lambda symbol: {
            "AAPL": {"marketCap": 3000000000000},  # $3T
            "GOOGL": {"marketCap": 2000000000000},  # $2T
            "MSFT": {"marketCap": 2500000000000},  # $2.5T
        }.get(symbol, {})

        manager = AllocatorManager(
            data_fetcher=mock_data_fetcher,
            backtest_runner=mock_backtest_runner,
            logging_manager=mock_logging_manager,
        )

        result = manager.calculate_market_cap_quantities(sample_config, sample_tickers)

        # Total market cap: $7.5T
        # AAPL weight: 3/7.5 = 0.4, allocation: $40,000
        # GOOGL weight: 2/7.5 = 0.267, allocation: $26,667
        # MSFT weight: 2.5/7.5 = 0.333, allocation: $33,333
        assert "AAPL" in result
        assert "GOOGL" in result
        assert "MSFT" in result

    def test_calculate_dynamic_quantities(
        self, mock_data_fetcher, mock_backtest_runner, mock_logging_manager, sample_config, sample_tickers
    ):
        """Test dynamic quantity calculation."""
        manager = AllocatorManager(
            data_fetcher=mock_data_fetcher,
            backtest_runner=mock_backtest_runner,
            logging_manager=mock_logging_manager,
        )

        result = manager.calculate_dynamic_quantities(sample_config, sample_tickers)

        # AAPL: 50% of $100,000 = $50,000 / $150 = 333.33 shares
        # GOOGL: 30% of $100,000 = $30,000 / $2800 = 10.71 shares
        # MSFT: 20% of $100,000 = $20,000 / $300 = 66.67 shares
        assert result["AAPL"] == pytest.approx(333.33, rel=0.01)
        assert result["GOOGL"] == pytest.approx(10.71, rel=0.01)
        assert result["MSFT"] == pytest.approx(66.67, rel=0.01)

    def test_calculate_auto_allocation_quantities(
        self, mock_data_fetcher, mock_backtest_runner, mock_logging_manager, sample_config, sample_tickers
    ):
        """Test auto allocation quantity calculation."""
        manager = AllocatorManager(
            data_fetcher=mock_data_fetcher,
            backtest_runner=mock_backtest_runner,
            logging_manager=mock_logging_manager,
        )

        result = manager.calculate_auto_allocation_quantities(sample_config, sample_tickers)

        # All tickers are TECH, and TECH has 60% allocation
        # Target capital: $100,000 * 0.95 = $95,000
        # TECH allocation: $95,000 * 0.6 = $57,000
        # Split equally among 3 TECH stocks: $19,000 each
        assert "AAPL" in result
        assert "GOOGL" in result
        assert "MSFT" in result

    def test_calculate_backtest_optimized_quantities(
        self, mock_data_fetcher, mock_backtest_runner, mock_logging_manager, sample_config, sample_tickers
    ):
        """Test backtest optimized quantity calculation."""
        manager = AllocatorManager(
            data_fetcher=mock_data_fetcher,
            backtest_runner=mock_backtest_runner,
            logging_manager=mock_logging_manager,
        )

        # Mock the backtest allocator's calculate_quantities method
        expected_result = {"AAPL": 300, "GOOGL": 10, "MSFT": 100}
        manager.backtest_allocator.calculate_quantities = MagicMock(return_value=expected_result)

        result = manager.calculate_backtest_optimized_quantities(sample_config, sample_tickers)

        assert result == expected_result
        manager.backtest_allocator.calculate_quantities.assert_called_once_with(sample_config, sample_tickers)

    def test_get_calculation_prices(self, mock_data_fetcher, mock_backtest_runner, mock_logging_manager, sample_config):
        """Test getting calculation prices."""
        manager = AllocatorManager(
            data_fetcher=mock_data_fetcher,
            backtest_runner=mock_backtest_runner,
            logging_manager=mock_logging_manager,
        )

        symbols = ["AAPL", "GOOGL", "MSFT"]
        result = manager.get_calculation_prices(sample_config, symbols, use_start_date=True)

        # Should use the mocked prices
        assert result == {"AAPL": 150.0, "GOOGL": 2800.0, "MSFT": 300.0}

    def test_validate_allocation_config_backtest_optimized(
        self, mock_data_fetcher, mock_backtest_runner, mock_logging_manager
    ):
        """Test validation for backtest optimized allocation."""
        manager = AllocatorManager(
            data_fetcher=mock_data_fetcher,
            backtest_runner=mock_backtest_runner,
            logging_manager=mock_logging_manager,
        )

        config = MagicMock()
        config.portfolio = MagicMock()
        config.portfolio.allocation_method = "backtest_optimized"
        config.backtest_optimization = None

        with pytest.raises(ValueError, match="backtest_optimization configuration is required"):
            manager.validate_allocation_config(config)

    def test_validate_allocation_config_auto(self, mock_data_fetcher, mock_backtest_runner, mock_logging_manager):
        """Test validation for auto allocation."""
        manager = AllocatorManager(
            data_fetcher=mock_data_fetcher,
            backtest_runner=mock_backtest_runner,
            logging_manager=mock_logging_manager,
        )

        config = MagicMock()
        config.portfolio = MagicMock()
        config.portfolio.allocation_method = "auto"
        config.portfolio.category_ratios = None

        with pytest.raises(ValueError, match="category_ratios must be specified"):
            manager.validate_allocation_config(config)

    def test_validate_allocation_config_dynamic(self, mock_data_fetcher, mock_backtest_runner, mock_logging_manager):
        """Test validation for dynamic allocation."""
        manager = AllocatorManager(
            data_fetcher=mock_data_fetcher,
            backtest_runner=mock_backtest_runner,
            logging_manager=mock_logging_manager,
        )

        config = MagicMock()
        config.portfolio = MagicMock()
        config.portfolio.allocation_method = "dynamic"
        config.portfolio.tickers = [
            TickerConfig(symbol="AAPL"),  # No allocation_pct or allocation_amount
            TickerConfig(symbol="GOOGL", allocation_pct=50),
        ]

        with pytest.raises(ValueError, match="Ticker AAPL must have either allocation_pct"):
            manager.validate_allocation_config(config)



"""Tests for domain models."""

from datetime import date
from unittest.mock import Mock, patch

import pandas as pd
import pytest

from stockula.allocation import Allocator
from stockula.config import DataConfig, PortfolioConfig, StockulaConfig, TickerConfig
from stockula.domain import Asset, Category, DomainFactory, Portfolio, Ticker, TickerRegistry


@pytest.fixture
def mock_logging_manager():
    """Create a mock logging manager."""
    logger = Mock()
    logger.debug = Mock()
    logger.info = Mock()
    logger.warning = Mock()
    logger.error = Mock()
    return logger


@pytest.fixture
def mock_allocator(mock_data_fetcher, mock_logging_manager):
    """Create a mock allocator with mocked logging manager."""
    allocator = Allocator(mock_data_fetcher, logging_manager=mock_logging_manager)
    return allocator


class TestTicker:
    """Test Ticker domain model."""

    def test_ticker_creation(self, mock_logging_manager):
        """Test creating a ticker."""
        ticker = Ticker(
            symbol="AAPL",
            sector="Technology",
            market_cap=3000.0,
            category=Category.MOMENTUM,
        )
        assert ticker.symbol == "AAPL"
        assert ticker.sector == "Technology"
        assert ticker.market_cap == 3000.0
        assert ticker.category == Category.MOMENTUM

    def test_ticker_defaults(self):
        """Test ticker with default values."""
        ticker = Ticker(symbol="GOOGL")
        assert ticker.symbol == "GOOGL"
        assert ticker.sector is None
        assert ticker.market_cap is None
        assert ticker.category is None

    def test_ticker_equality(self):
        """Test ticker equality based on symbol."""
        ticker1 = Ticker(symbol="AAPL", sector="Technology")
        ticker2 = Ticker(symbol="AAPL", sector="Tech")  # Different sector
        ticker3 = Ticker(symbol="GOOGL")

        assert ticker1 == ticker2  # Same symbol
        assert ticker1 != ticker3  # Different symbol

    def test_ticker_hash(self):
        """Test ticker hashing for use in sets/dicts."""
        ticker1 = Ticker(symbol="AAPL")
        ticker2 = Ticker(symbol="AAPL")
        ticker3 = Ticker(symbol="GOOGL")

        ticker_set = {ticker1, ticker2, ticker3}
        assert len(ticker_set) == 2  # ticker1 and ticker2 are the same

    def test_ticker_string_representation(self):
        """Test ticker string representation."""
        ticker = Ticker(symbol="AAPL", sector="Technology")
        assert str(ticker) == "Ticker(AAPL)"
        assert repr(ticker) == "Ticker(symbol='AAPL', sector='Technology', market_cap=None, category=None)"


class TestTickerRegistry:
    """Test TickerRegistry singleton."""

    def test_ticker_registry_singleton(self):
        """Test that TickerRegistry is a singleton."""
        registry1 = TickerRegistry()
        registry2 = TickerRegistry()
        assert registry1 is registry2

    def test_get_or_create_ticker(self):
        """Test get_or_create ticker functionality."""
        registry = TickerRegistry()

        # Create new ticker
        ticker1 = registry.get_or_create("AAPL", sector="Technology")
        assert ticker1.symbol == "AAPL"
        assert ticker1.sector == "Technology"

        # Get existing ticker with different sector (should update the ticker)
        ticker2 = registry.get_or_create("AAPL", sector="Tech")  # Different sector
        assert ticker2.symbol == "AAPL"
        assert ticker2.sector == "Tech"  # Updated sector

        # Get without providing sector should keep current value
        ticker3 = registry.get_or_create("AAPL")
        assert ticker3.sector == "Tech"

    def test_get_ticker(self):
        """Test getting a ticker from registry."""
        registry = TickerRegistry()

        # Ticker doesn't exist
        assert registry.get("MSFT") is None

        # Create ticker
        ticker = registry.get_or_create("MSFT")

        # Now it exists
        assert registry.get("MSFT") is ticker

    def test_all_tickers(self):
        """Test getting all tickers from registry."""
        registry = TickerRegistry()

        # Create multiple tickers
        ticker1 = registry.get_or_create("AAPL")
        registry.get_or_create("GOOGL")
        registry.get_or_create("MSFT")

        all_tickers = registry.all()
        assert len(all_tickers) == 3
        assert "AAPL" in all_tickers
        assert all_tickers["AAPL"] is ticker1

    def test_clear_registry(self):
        """Test clearing the ticker registry."""
        registry = TickerRegistry()

        # Add some tickers
        registry.get_or_create("AAPL")
        registry.get_or_create("GOOGL")
        assert len(registry.all()) == 2

        # Clear registry
        registry._clear()
        assert len(registry.all()) == 0


class TestAsset:
    """Test Asset domain model."""

    def test_asset_creation(self, sample_ticker):
        """Test creating an asset."""
        asset = Asset(ticker_init=sample_ticker, quantity_init=10.0, category_init=Category.MOMENTUM)
        assert asset.ticker == sample_ticker
        assert asset.quantity == 10.0
        assert asset.category == Category.MOMENTUM
        assert asset.symbol == "AAPL"

    def test_asset_without_category(self):
        """Test asset without explicit category is None."""
        ticker_with_category = Ticker(
            "NVDA",  # symbol_init
            None,  # sector_init
            None,  # market_cap_init
            "SPECULATIVE",  # category_init
            None,  # price_range_init
            None,  # metadata_init
        )
        asset = Asset(ticker_init=ticker_with_category, quantity_init=5.0, category_init=None)
        assert asset.category is None  # Asset doesn't inherit ticker category

    def test_asset_category_override(self):
        """Test asset can have its own category."""
        ticker_with_category = Ticker(
            "SPY",  # symbol_init
            None,  # sector_init
            None,  # market_cap_init
            "INDEX",  # category_init
            None,  # price_range_init
            None,  # metadata_init
        )
        asset = Asset(
            ticker_init=ticker_with_category,
            quantity_init=20.0,
            category_init=Category.GROWTH,  # Asset's own category
        )
        assert asset.category == Category.GROWTH
        assert ticker_with_category.category == "INDEX"  # Ticker unchanged

    def test_asset_value_calculation(self, sample_ticker):
        """Test asset value calculation."""
        asset = Asset(ticker_init=sample_ticker, quantity_init=10.0)

        # Test with different prices
        assert asset.get_value(150.0) == 1500.0
        assert asset.get_value(200.0) == 2000.0
        assert asset.get_value(0.0) == 0.0

    def test_asset_string_representation(self, sample_ticker):
        """Test asset string representation."""
        asset = Asset(ticker_init=sample_ticker, quantity_init=10.0)
        # Asset.__str__ returns: Asset(symbol, quantity shares[, category])
        assert "AAPL" in str(asset)
        assert "10.00 shares" in str(asset)


class TestCategory:
    """Test Category enum."""

    def test_category_values(self):
        """Test category enum values."""
        # Check that categories exist and have proper string representations
        assert str(Category.INDEX) == "Index"
        assert str(Category.LARGE_CAP) == "Large Cap"
        assert str(Category.MOMENTUM) == "Momentum"
        assert str(Category.GROWTH) == "Growth"
        assert str(Category.VALUE) == "Value"
        assert str(Category.DIVIDEND) == "Dividend"
        assert str(Category.SPECULATIVE) == "Speculative"
        assert str(Category.INTERNATIONAL) == "International"
        assert str(Category.COMMODITY) == "Commodity"
        assert str(Category.BOND) == "Bond"
        assert str(Category.CRYPTO) == "Crypto"

        # Test that values are integers (from auto())
        assert isinstance(Category.INDEX.value, int)
        assert isinstance(Category.GROWTH.value, int)

    def test_category_from_string(self):
        """Test creating category from string."""
        assert Category["INDEX"] == Category.INDEX
        assert Category["MOMENTUM"] == Category.MOMENTUM

        with pytest.raises(KeyError):
            Category["INVALID_CATEGORY"]


class TestPortfolio:
    """Test Portfolio domain model."""

    def test_portfolio_creation(self, mock_logging_manager):
        """Test creating a portfolio."""
        portfolio = Portfolio(
            name_init="Test Portfolio",
            initial_capital_init=100000.0,
            allocation_method_init="equal_weight",
            logging_manager_init=mock_logging_manager,
            max_position_size=25.0,
            stop_loss_pct=10.0,
        )
        assert portfolio.name == "Test Portfolio"
        assert portfolio.initial_capital == 100000.0
        assert portfolio.allocation_method == "equal_weight"
        assert portfolio.max_position_size == 25.0
        assert portfolio.stop_loss_pct == 10.0
        assert len(portfolio.assets) == 0

    def test_add_asset(self, sample_portfolio, sample_asset):
        """Test adding an asset to portfolio."""
        sample_portfolio.add_asset(sample_asset)
        assert len(sample_portfolio.assets) == 1
        assert sample_portfolio.assets[0] == sample_asset

    def test_add_duplicate_asset_raises_error(self, sample_portfolio, sample_asset):
        """Test adding duplicate asset raises error."""
        sample_portfolio.add_asset(sample_asset)

        # Try to add same ticker again
        duplicate_asset = Asset(ticker_init=sample_asset.ticker, quantity_init=5.0)
        with pytest.raises(ValueError, match="already exists"):
            sample_portfolio.add_asset(duplicate_asset)

    def test_get_all_assets(self, populated_portfolio):
        """Test getting all assets from portfolio."""
        assets = populated_portfolio.get_all_assets()
        assert len(assets) == 4
        symbols = [asset.symbol for asset in assets]
        assert "AAPL" in symbols
        assert "GOOGL" in symbols
        assert "SPY" in symbols
        assert "NVDA" in symbols

    def test_get_assets_by_category(self, populated_portfolio):
        """Test getting assets by category."""
        momentum_assets = populated_portfolio.get_assets_by_category(Category.MOMENTUM)
        assert len(momentum_assets) == 1
        assert momentum_assets[0].symbol == "AAPL"

        index_assets = populated_portfolio.get_assets_by_category(Category.INDEX)
        assert len(index_assets) == 1
        assert index_assets[0].symbol == "SPY"

    def test_get_asset_by_symbol(self, populated_portfolio):
        """Test getting asset by symbol."""
        asset = populated_portfolio.get_asset_by_symbol("AAPL")
        assert asset is not None
        assert asset.symbol == "AAPL"

        # Non-existent symbol
        assert populated_portfolio.get_asset_by_symbol("TSLA") is None

    def test_portfolio_value_calculation(self, populated_portfolio, sample_prices):
        """Test portfolio value calculation."""
        value = populated_portfolio.get_portfolio_value(sample_prices)

        # Calculate expected value
        expected = (
            10.0 * 150.0  # AAPL
            + 5.0 * 120.0  # GOOGL
            + 20.0 * 450.0  # SPY
            + 8.0 * 500.0
        )  # NVDA
        assert value == expected

    def test_portfolio_value_with_missing_prices(self, populated_portfolio):
        """Test portfolio value calculation with missing prices."""
        partial_prices = {"AAPL": 150.0, "GOOGL": 120.0}  # Missing SPY and NVDA

        # Should calculate value for assets with prices, skip others
        value = populated_portfolio.get_portfolio_value(partial_prices)
        expected = 10.0 * 150.0 + 5.0 * 120.0  # Only AAPL and GOOGL
        assert value == expected

    def test_get_allocation_by_category(self, populated_portfolio, sample_prices):
        """Test getting allocation by category."""
        allocations = populated_portfolio.get_allocation_by_category(sample_prices)

        # Check that category names are in allocations
        assert "Momentum" in allocations
        assert "Index" in allocations
        assert "Growth" in allocations
        assert "Speculative" in allocations

        # Check percentages sum to 100
        total_pct = sum(alloc["percentage"] for alloc in allocations.values())
        assert abs(total_pct - 100.0) < 0.01

        # Check specific category
        momentum_alloc = allocations["Momentum"]
        assert momentum_alloc["value"] == 10.0 * 150.0  # AAPL
        assert momentum_alloc["assets"] == ["AAPL"]

    def test_validate_capital_sufficiency(self, sample_portfolio, sample_asset):
        """Test capital sufficiency validation."""
        # Add assets that exceed capital
        expensive_ticker = Ticker(
            "BRK.A",  # symbol_init
            None,  # sector_init
            None,  # market_cap_init
            "VALUE",  # category_init
            None,  # price_range_init
            None,  # metadata_init
        )
        expensive_asset = Asset(ticker_init=expensive_ticker, quantity_init=1.0, category_init=Category.VALUE)

        sample_portfolio.add_asset(sample_asset)
        sample_portfolio.add_asset(expensive_asset)

        # Mock prices
        with patch.object(sample_portfolio, "get_portfolio_value") as mock_value:
            mock_value.return_value = 150000.0  # Exceeds initial capital of 100000

            with pytest.raises(ValueError, match="is insufficient to cover"):
                sample_portfolio.validate_capital_sufficiency()

    def test_validate_allocation_constraints(self, sample_portfolio, sample_asset):
        """Test allocation constraints validation."""
        sample_portfolio.max_position_size = 20.0  # 20% max
        sample_portfolio.add_asset(sample_asset)

        # Mock to make one position exceed 20%
        mock_allocations = {
            "AAPL": {
                "value": 250.0,
                "percentage": 25.0,  # Exceeds 20% max
                "quantity": 10.0,
            }
        }

        with patch.object(sample_portfolio, "get_asset_allocations") as mock_alloc:
            mock_alloc.return_value = mock_allocations

            # Should raise error for exceeding max position size
            with pytest.raises(ValueError, match="exceeds maximum position size"):
                sample_portfolio.validate_allocation_constraints()


class TestDomainFactory:
    """Test DomainFactory."""

    def test_create_portfolio_basic(self, sample_stockula_config, mock_data_fetcher, mock_logging_manager):
        """Test creating a basic portfolio from config."""
        factory = DomainFactory(fetcher=mock_data_fetcher, logging_manager=mock_logging_manager)

        # Mock validation methods to not raise errors
        with patch.object(Portfolio, "validate_capital_sufficiency"):
            with patch.object(Portfolio, "validate_allocation_constraints"):
                portfolio = factory.create_portfolio(sample_stockula_config)

        assert portfolio.name == "Test Portfolio"
        assert portfolio.initial_capital == 100000.0
        assert len(portfolio.assets) == 4

    def test_create_portfolio_with_dynamic_allocation(
        self, dynamic_allocation_config, mock_data_fetcher, mock_allocator, mock_logging_manager
    ):
        """Test creating portfolio with dynamic allocation."""
        config = StockulaConfig(portfolio=dynamic_allocation_config)
        factory = DomainFactory(
            fetcher=mock_data_fetcher, allocator_manager=mock_allocator, logging_manager=mock_logging_manager
        )

        portfolio = factory.create_portfolio(config)

        assert len(portfolio.assets) == 3

        # Check AAPL allocation (fixed $15,000 / $150 = 100 shares)
        aapl_asset = portfolio.get_asset_by_symbol("AAPL")
        assert aapl_asset.quantity == 100.0

        # Check GOOGL allocation (20% of $50,000 = $10,000 / $120 ≈ 83.33 shares)
        googl_asset = portfolio.get_asset_by_symbol("GOOGL")
        assert googl_asset.quantity == pytest.approx(83.33, rel=0.01)

    def test_create_portfolio_with_auto_allocation(
        self, auto_allocation_config, mock_data_fetcher, mock_allocator, mock_logging_manager
    ):
        """Test creating portfolio with auto allocation."""
        config = StockulaConfig(portfolio=auto_allocation_config)
        factory = DomainFactory(
            fetcher=mock_data_fetcher, allocator_manager=mock_allocator, logging_manager=mock_logging_manager
        )

        portfolio = factory.create_portfolio(config)

        assert len(portfolio.assets) == 5

        # Check that all assets have quantities
        for asset in portfolio.assets:
            assert asset.quantity > 0

        # Check category allocation ratios are roughly maintained
        allocations = portfolio.get_allocation_by_category(
            mock_data_fetcher.get_current_prices([a.symbol for a in portfolio.assets])
        )

        # INDEX should be roughly 35%
        index_pct = allocations["Index"]["percentage"]
        assert 30 < index_pct < 40  # Allow some variance

    def test_create_portfolio_dynamic_allocation_without_info(self, mock_logging_manager):
        """Test creating portfolio with dynamic allocation but no allocation info."""
        # This is now allowed - the ticker config doesn't validate against portfolio settings
        config = StockulaConfig(
            portfolio=PortfolioConfig(
                name="Test",
                initial_capital=10000,
                dynamic_allocation=True,
                tickers=[
                    TickerConfig(symbol="AAPL"),  # No allocation info
                ],
            )
        )
        # The validation would happen later when trying to calculate dynamic quantities
        assert config.portfolio.dynamic_allocation is True
        assert config.portfolio.tickers[0].quantity is None

    def test_get_all_tickers(self, sample_stockula_config, mock_data_fetcher, mock_logging_manager):
        """Test getting all tickers from factory."""
        factory = DomainFactory(fetcher=mock_data_fetcher, logging_manager=mock_logging_manager)

        # Create portfolio to populate registry
        with patch.object(Portfolio, "validate_capital_sufficiency"):
            with patch.object(Portfolio, "validate_allocation_constraints"):
                factory.create_portfolio(sample_stockula_config)

        all_tickers = factory.get_all_tickers()
        assert len(all_tickers) == 4
        symbols = [t.symbol for t in all_tickers]
        assert "AAPL" in symbols
        assert "GOOGL" in symbols


class TestDomainFactoryAdvanced:
    """Test advanced DomainFactory functionality."""

    def test_create_portfolio_edge_cases(self, mock_data_fetcher, mock_logging_manager):
        """Test portfolio creation edge cases."""
        factory = DomainFactory(fetcher=mock_data_fetcher, logging_manager=mock_logging_manager)

        # Test with minimal config
        minimal_config = StockulaConfig(
            portfolio=PortfolioConfig(
                name="Minimal",
                initial_capital=1000.0,
                allocation_method="equal_weight",
                tickers=[TickerConfig(symbol="SPY", quantity=1.0, category="INDEX")],
            )
        )

        with patch.object(Portfolio, "validate_capital_sufficiency"):
            with patch.object(Portfolio, "validate_allocation_constraints"):
                portfolio = factory.create_portfolio(minimal_config)

        assert portfolio.name == "Minimal"
        assert portfolio.initial_capital == 1000.0
        assert len(portfolio.assets) == 1

    def test_create_portfolio_with_categories(self, mock_data_fetcher, mock_logging_manager):
        """Test portfolio creation with various category types."""
        config = StockulaConfig(
            portfolio=PortfolioConfig(
                name="Category Test",
                initial_capital=100000.0,
                allocation_method="equal_weight",
                tickers=[
                    TickerConfig(symbol="VTI", quantity=10.0, category="INDEX"),
                    TickerConfig(symbol="QQQ", quantity=5.0, category="GROWTH"),
                    TickerConfig(symbol="VYM", quantity=15.0, category="DIVIDEND"),
                    TickerConfig(symbol="ARKK", quantity=8.0, category="SPECULATIVE"),
                ],
            )
        )

        factory = DomainFactory(fetcher=mock_data_fetcher, logging_manager=mock_logging_manager)

        with patch.object(Portfolio, "validate_capital_sufficiency"):
            with patch.object(Portfolio, "validate_allocation_constraints"):
                portfolio = factory.create_portfolio(config)

        # Check categories are correctly assigned
        categories = {asset.category for asset in portfolio.assets}
        expected_categories = {
            Category.INDEX,
            Category.GROWTH,
            Category.DIVIDEND,
            Category.SPECULATIVE,
        }
        assert categories == expected_categories

    def test_create_portfolio_dynamic_allocation_error_handling(
        self, mock_data_fetcher, mock_allocator, mock_logging_manager
    ):
        """Test error handling in dynamic allocation."""
        # Patch the get_current_prices method to raise an exception
        with patch.object(
            mock_data_fetcher,
            "get_current_prices",
            side_effect=Exception("Price fetch failed"),
        ):
            config = StockulaConfig(
                portfolio=PortfolioConfig(
                    name="Error Test",
                    initial_capital=10000.0,
                    dynamic_allocation=True,
                    tickers=[TickerConfig(symbol="AAPL", allocation_amount=5000.0)],
                )
            )

            factory = DomainFactory(
                fetcher=mock_data_fetcher, allocator_manager=mock_allocator, logging_manager=mock_logging_manager
            )

            # Should handle price fetch errors gracefully or raise appropriate error
            with pytest.raises(Exception) as exc_info:
                factory.create_portfolio(config)
            assert "Price fetch failed" in str(exc_info.value)

    def test_create_portfolio_auto_allocation_categories(self, mock_data_fetcher, mock_allocator, mock_logging_manager):
        """Test auto allocation with category ratios."""
        # Override the return value for this specific test
        with patch.object(
            mock_data_fetcher,
            "get_current_prices",
            return_value={
                "VTI": 200.0,
                "QQQ": 300.0,
                "VYM": 100.0,
                "ARKK": 50.0,
                "VWO": 45.0,
            },
        ):
            config = StockulaConfig(
                portfolio=PortfolioConfig(
                    name="Auto Allocation",
                    initial_capital=100000.0,
                    auto_allocate=True,
                    category_ratios={
                        "INDEX": 0.4,  # 40%
                        "GROWTH": 0.3,  # 30%
                        "DIVIDEND": 0.2,  # 20%
                        "SPECULATIVE": 0.1,  # 10%
                        "INTERNATIONAL": 0.0,  # 0% for VWO
                    },
                    tickers=[
                        TickerConfig(symbol="VTI", category="INDEX"),
                        TickerConfig(symbol="QQQ", category="GROWTH"),
                        TickerConfig(symbol="VYM", category="DIVIDEND"),
                        TickerConfig(symbol="ARKK", category="SPECULATIVE"),
                        TickerConfig(symbol="VWO", category="INTERNATIONAL"),
                    ],
                )
            )

            factory = DomainFactory(
                fetcher=mock_data_fetcher, allocator_manager=mock_allocator, logging_manager=mock_logging_manager
            )
            portfolio = factory.create_portfolio(config)

            # Only tickers with non-zero category allocation should be included
            # VWO has 0% allocation for INTERNATIONAL category, so it's excluded
            assert len(portfolio.assets) == 4

            # Check that quantities were calculated
            for asset in portfolio.assets:
                assert asset.quantity > 0
                # VWO should not be in the portfolio
                assert asset.ticker.symbol != "VWO"

    def test_ticker_registry_integration(self, mock_data_fetcher, mock_logging_manager):
        """Test ticker registry integration with factory."""
        factory = DomainFactory(fetcher=mock_data_fetcher, logging_manager=mock_logging_manager)

        # Clear registry first
        registry = TickerRegistry()
        registry._clear()

        config = StockulaConfig(
            portfolio=PortfolioConfig(
                name="Registry Test",
                initial_capital=50000.0,
                allocation_method="equal_weight",
                tickers=[
                    TickerConfig(symbol="AAPL", quantity=5.0),
                    TickerConfig(symbol="GOOGL", quantity=3.0),
                ],
            )
        )

        with patch.object(Portfolio, "validate_capital_sufficiency"):
            with patch.object(Portfolio, "validate_allocation_constraints"):
                factory.create_portfolio(config)

        # Check that tickers were added to registry
        all_tickers = factory.get_all_tickers()
        assert len(all_tickers) == 2

        # Check registry contains the tickers
        registry_tickers = registry.all()
        assert "AAPL" in registry_tickers
        assert "GOOGL" in registry_tickers

    def test_factory_methods_consistency(self, mock_data_fetcher, mock_logging_manager):
        """Test that factory methods are consistent."""
        factory = DomainFactory(fetcher=mock_data_fetcher, logging_manager=mock_logging_manager)

        # Test that get_all_tickers returns empty list initially
        initial_tickers = factory.get_all_tickers()
        assert isinstance(initial_tickers, list)
        assert len(initial_tickers) >= 0  # Could have tickers from other tests

    def test_allocation_amount_calculation(self, mock_data_fetcher, mock_allocator, mock_logging_manager):
        """Test allocation amount calculation accuracy."""
        # Override the return value for this specific test
        # Create a mock allocator that returns the expected quantities
        mock_allocator_manager = Mock()
        mock_allocator_manager.calculate_quantities.return_value = {"AAPL": 100.0, "GOOGL": 2.0}

        config = StockulaConfig(
            portfolio=PortfolioConfig(
                name="Allocation Test",
                initial_capital=100000.0,
                dynamic_allocation=True,
                allow_fractional_shares=True,  # Need this for exact allocation
                tickers=[
                    TickerConfig(symbol="AAPL", allocation_amount=15000.0),  # $15k / $150 = 100 shares
                    TickerConfig(symbol="GOOGL", allocation_amount=5000.0),  # $5k / $2500 = 2 shares
                ],
            )
        )

        factory = DomainFactory(
            fetcher=mock_data_fetcher, allocator_manager=mock_allocator_manager, logging_manager=mock_logging_manager
        )
        portfolio = factory.create_portfolio(config)

        # Check exact allocations
        aapl_asset = portfolio.get_asset_by_symbol("AAPL")
        assert aapl_asset.quantity == 100.0

        googl_asset = portfolio.get_asset_by_symbol("GOOGL")
        assert googl_asset.quantity == 2.0

    def test_allocation_percentage_calculation(self, mock_data_fetcher, mock_allocator, mock_logging_manager):
        """Test allocation percentage calculation."""
        # Override the return value for this specific test
        # Create a mock allocator that returns the expected quantities
        mock_allocator_manager = Mock()
        mock_allocator_manager.calculate_quantities.return_value = {"VTI": 180.0, "QQQ": 80.0}

        config = StockulaConfig(
            portfolio=PortfolioConfig(
                name="Percentage Test",
                initial_capital=60000.0,
                dynamic_allocation=True,
                allow_fractional_shares=True,  # Need this for exact allocation
                tickers=[
                    TickerConfig(symbol="VTI", allocation_pct=60.0),  # 60% of $60k = $36k / $200 = 180 shares
                    TickerConfig(symbol="QQQ", allocation_pct=40.0),  # 40% of $60k = $24k / $300 = 80 shares
                ],
            )
        )

        factory = DomainFactory(
            fetcher=mock_data_fetcher, allocator_manager=mock_allocator_manager, logging_manager=mock_logging_manager
        )
        portfolio = factory.create_portfolio(config)

        vti_asset = portfolio.get_asset_by_symbol("VTI")
        assert vti_asset.quantity == 180.0

        qqq_asset = portfolio.get_asset_by_symbol("QQQ")
        assert qqq_asset.quantity == 80.0


class TestDomainFactoryErrorScenarios:
    """Test error scenarios in DomainFactory."""

    def test_create_portfolio_zero_capital(self, mock_logging_manager):
        """Test portfolio creation with zero capital."""
        from pydantic import ValidationError

        # Zero capital should be caught by pydantic validation
        with pytest.raises(ValidationError):
            StockulaConfig(
                portfolio=PortfolioConfig(
                    name="Zero Capital",
                    initial_capital=0.0,  # This should fail validation
                    allocation_method="equal_weight",
                    tickers=[TickerConfig(symbol="SPY", quantity=1.0)],
                )
            )

    def test_create_portfolio_zero_price(self, mock_data_fetcher, mock_allocator, mock_logging_manager):
        """Test handling of zero stock prices."""
        # Override the return value for this specific test
        with patch.object(
            mock_data_fetcher,
            "get_current_prices",
            return_value={
                "PENNY": 0.0  # Zero price
            },
        ):
            config = StockulaConfig(
                portfolio=PortfolioConfig(
                    name="Zero Price Test",
                    initial_capital=10000.0,
                    dynamic_allocation=True,
                    tickers=[TickerConfig(symbol="PENNY", allocation_amount=1000.0)],
                )
            )

            factory = DomainFactory(
                fetcher=mock_data_fetcher, allocator_manager=mock_allocator, logging_manager=mock_logging_manager
            )

            # Should handle zero price gracefully or raise appropriate error
            with pytest.raises((ValueError, ZeroDivisionError)):
                factory.create_portfolio(config)

    def test_create_portfolio_missing_prices(self, mock_data_fetcher, mock_logging_manager):
        """Test handling of missing stock prices."""
        # Override the return value for this specific test
        with patch.object(
            mock_data_fetcher,
            "get_current_prices",
            return_value={
                "AAPL": 150.0
                # GOOGL missing
            },
        ):
            config = StockulaConfig(
                portfolio=PortfolioConfig(
                    name="Missing Price Test",
                    initial_capital=10000.0,
                    dynamic_allocation=True,
                    tickers=[
                        TickerConfig(symbol="AAPL", allocation_amount=5000.0),
                        TickerConfig(symbol="GOOGL", allocation_amount=5000.0),  # Price missing
                    ],
                )
            )

            # Create a mock allocator_manager that will raise the expected error
            mock_allocator_manager = Mock()
            mock_allocator_manager.calculate_quantities.side_effect = ValueError("Could not fetch price for GOOGL")

            factory = DomainFactory(
                fetcher=mock_data_fetcher,
                allocator_manager=mock_allocator_manager,
                logging_manager=mock_logging_manager,
            )

            # Should raise error for missing prices in dynamic allocation
            with pytest.raises(ValueError, match="Could not fetch price for GOOGL"):
                factory.create_portfolio(config)


class TestTickerRegistryAdvanced:
    """Test advanced TickerRegistry functionality."""

    def test_ticker_registry_update_behavior(self, mock_logging_manager):
        """Test ticker registry update behavior."""
        registry = TickerRegistry()
        registry._clear()

        # Create ticker with initial data - use all required parameters
        ticker1 = registry.get_or_create(
            "AAPL",
            sector="Technology",
            market_cap=3000.0,
            category=None,
            price_range=None,
        )
        assert ticker1.sector == "Technology"
        assert ticker1.market_cap == 3000.0

        # Update with new data - use all required parameters
        ticker2 = registry.get_or_create("AAPL", sector="Tech", market_cap=3100.0, category=None, price_range=None)
        # Since Ticker is immutable, a new instance is created with updated values
        assert ticker2 is not ticker1  # Different object due to immutability
        assert ticker2.symbol == ticker1.symbol  # Same symbol
        assert ticker2.sector == "Tech"  # Updated
        assert ticker2.market_cap == 3100.0  # Updated

        # But the registry returns the new instance for the same symbol
        ticker3 = registry.get("AAPL")
        assert ticker3 is ticker2

    def test_ticker_registry_category_handling(self):
        """Test ticker registry category handling."""
        registry = TickerRegistry()
        registry._clear()

        # Create ticker with category
        ticker = registry.get_or_create("NVDA", category=Category.SPECULATIVE)
        assert ticker.category == Category.SPECULATIVE

        # Update category
        updated_ticker = registry.get_or_create("NVDA", category=Category.GROWTH)
        assert updated_ticker.category == Category.GROWTH

    def test_ticker_registry_concurrent_access(self):
        """Test ticker registry handles concurrent-like access."""
        registry = TickerRegistry()
        registry._clear()

        # Simulate multiple "concurrent" creations
        tickers = []
        for i in range(10):
            ticker = registry.get_or_create(f"STOCK{i}")
            tickers.append(ticker)

        # All should be unique
        assert len(tickers) == 10
        assert len({ticker.symbol for ticker in tickers}) == 10

        # Registry should contain all
        all_tickers = registry.all()
        assert len(all_tickers) == 10


class TestAssetAdvanced:
    """Test advanced Asset functionality."""

    def test_asset_value_edge_cases(self):
        """Test asset value calculation edge cases."""
        ticker = Ticker("TEST")
        asset = Asset(ticker_init=ticker, quantity_init=10.0)

        # Test with negative price (should handle gracefully)
        value = asset.get_value(-50.0)
        assert value == -500.0  # Mathematically correct

        # Test with very large numbers
        large_value = asset.get_value(1e10)
        assert large_value == 1e11

        # Test with very small numbers
        small_value = asset.get_value(0.0001)
        assert small_value == 0.001

    def test_asset_quantity_edge_cases(self):
        """Test asset with edge case quantities."""
        # Use proper Ticker constructor with all required parameters
        ticker = Ticker(
            "TEST",  # symbol_init
            None,  # sector_init
            None,  # market_cap_init
            None,  # category_init
            None,  # price_range_init
            None,  # metadata_init
        )

        # Fractional shares
        fractional_asset = Asset(ticker_init=ticker, quantity_init=0.5)
        assert fractional_asset.get_value(100.0) == 50.0

        # Assets must have positive quantity, so test minimum positive value
        min_asset = Asset(ticker_init=ticker, quantity_init=0.001)
        assert min_asset.get_value(100.0) == 0.1

        # Very large quantity
        large_asset = Asset(ticker_init=ticker, quantity_init=1e6)
        assert large_asset.get_value(100.0) == 1e8

    def test_asset_category_precedence(self):
        """Test asset category precedence over ticker category."""
        ticker_with_category = Ticker("TEST", category=Category.INDEX)

        # Asset category should override ticker category
        asset = Asset(ticker_init=ticker_with_category, quantity_init=10.0, category_init=Category.GROWTH)
        assert asset.category == Category.GROWTH
        assert ticker_with_category.category == Category.INDEX  # Unchanged


class TestDateToString:
    """Test the date_to_string utility function."""

    def test_date_to_string_with_none(self):
        """Test date_to_string with None input."""
        from stockula.domain.factory import date_to_string

        assert date_to_string(None) is None

    def test_date_to_string_with_string(self):
        """Test date_to_string with string input."""
        from stockula.domain.factory import date_to_string

        assert date_to_string("2023-01-01") == "2023-01-01"

    def test_date_to_string_with_date(self):
        """Test date_to_string with date object."""
        from stockula.domain.factory import date_to_string

        test_date = date(2023, 1, 1)
        assert date_to_string(test_date) == "2023-01-01"


class TestDomainFactoryEdgeCases:
    """Test edge cases in DomainFactory."""

    def test_create_portfolio_without_config(self, mock_logging_manager):
        """Test creating portfolio without config."""
        factory = DomainFactory(logging_manager=mock_logging_manager)

        # Create a config to pass
        config = StockulaConfig(
            portfolio=PortfolioConfig(
                name="Test",
                initial_capital=100000,
                tickers=[TickerConfig(symbol="AAPL", quantity=10)],
            )
        )

        # Without fetcher, should raise error for any allocation method requiring prices
        factory = DomainFactory(fetcher=None, logging_manager=mock_logging_manager)
        config.portfolio.dynamic_allocation = True
        config.portfolio.tickers[0].allocation_pct = 100.0

        with pytest.raises(ValueError, match="AllocatorManager not configured"):
            factory.create_portfolio(config)

    def test_calculate_dynamic_quantities_with_start_date_no_data(self, mock_allocator, mock_logging_manager):
        """Test quantity calculation when no data available on start date."""
        # Create config with start date
        config = StockulaConfig(
            data=DataConfig(start_date="2023-01-01", end_date="2023-12-31"),
            portfolio=PortfolioConfig(
                name="Test Portfolio",
                initial_capital=100000,
                dynamic_allocation=True,
                tickers=[
                    TickerConfig(symbol="AAPL", category="MOMENTUM", allocation_pct=50.0),
                    TickerConfig(symbol="GOOGL", category="GROWTH", allocation_pct=50.0),
                ],
            ),
        )

        # Mock fetcher
        mock_fetcher = Mock()

        # First call returns empty data (no data on exact date)
        # Second call returns data from a week later
        mock_fetcher.get_stock_data.side_effect = [
            pd.DataFrame(),  # Empty for AAPL on start date
            pd.DataFrame({"Close": [150.0]}, index=[pd.Timestamp("2023-01-08")]),  # AAPL week later
            pd.DataFrame(),  # Empty for GOOGL on start date
            pd.DataFrame({"Close": [120.0]}, index=[pd.Timestamp("2023-01-08")]),  # GOOGL week later
        ]

        mock_fetcher.get_current_prices.return_value = {"AAPL": 155.0, "GOOGL": 125.0}

        # Update the allocator's fetcher
        mock_allocator.fetcher = mock_fetcher

        factory = DomainFactory(
            config=config, fetcher=mock_fetcher, allocator_manager=mock_allocator, logging_manager=mock_logging_manager
        )

        # Calculate quantities
        quantities = factory.allocator_manager.calculate_dynamic_quantities(config, config.portfolio.tickers)

        # Should have used the week-later prices
        assert "AAPL" in quantities
        assert "GOOGL" in quantities
        assert quantities["AAPL"] > 0
        assert quantities["GOOGL"] > 0

    def test_calculate_dynamic_quantities_with_start_date_fallback_to_current(
        self, mock_allocator, mock_logging_manager
    ):
        """Test quantity calculation falling back to current prices."""
        # Create config with start date
        config = StockulaConfig(
            data=DataConfig(start_date="2023-01-01", end_date="2023-12-31"),
            portfolio=PortfolioConfig(
                name="Test Portfolio",
                initial_capital=100000,
                dynamic_allocation=True,
                tickers=[TickerConfig(symbol="AAPL", category="MOMENTUM", allocation_pct=100.0)],
            ),
        )

        # Mock fetcher - all historical data calls return empty
        mock_fetcher = Mock()
        mock_fetcher.get_stock_data.return_value = pd.DataFrame()  # Always empty
        mock_fetcher.get_current_prices.return_value = {"AAPL": 155.0}

        # Update the allocator's fetcher
        mock_allocator.fetcher = mock_fetcher

        factory = DomainFactory(
            config=config, fetcher=mock_fetcher, allocator_manager=mock_allocator, logging_manager=mock_logging_manager
        )

        # Calculate quantities
        quantities = factory.allocator_manager.calculate_dynamic_quantities(config, config.portfolio.tickers)

        # Should have used current price as fallback
        assert "AAPL" in quantities
        assert quantities["AAPL"] > 0

    def test_calculate_dynamic_quantities_with_exception_handling(self, mock_allocator, mock_logging_manager):
        """Test quantity calculation with exception in data fetching."""
        # Create config with start date as string
        config = StockulaConfig(
            data=DataConfig(start_date="2023-01-01", end_date="2023-12-31"),
            portfolio=PortfolioConfig(
                name="Test Portfolio",
                initial_capital=100000,
                dynamic_allocation=True,
                tickers=[TickerConfig(symbol="AAPL", category="MOMENTUM", allocation_pct=100.0)],
            ),
        )

        # Mock fetcher that raises exception
        mock_fetcher = Mock()
        mock_fetcher.get_stock_data.side_effect = Exception("API Error")
        mock_fetcher.get_current_prices.return_value = {"AAPL": 155.0}

        # Update the allocator's fetcher
        mock_allocator.fetcher = mock_fetcher

        factory = DomainFactory(
            config=config, fetcher=mock_fetcher, allocator_manager=mock_allocator, logging_manager=mock_logging_manager
        )

        # Calculate quantities - should handle exception and use current prices
        quantities = factory.allocator_manager.calculate_dynamic_quantities(config, config.portfolio.tickers)

        assert "AAPL" in quantities
        assert quantities["AAPL"] > 0

    def test_calculate_dynamic_quantities_date_type_handling(self, mock_allocator, mock_logging_manager):
        """Test quantity calculation with date object instead of string."""
        # Create config with start date as date object
        config = StockulaConfig(
            data=DataConfig(start_date=date(2023, 1, 1), end_date=date(2023, 12, 31)),
            portfolio=PortfolioConfig(
                name="Test Portfolio",
                initial_capital=100000,
                dynamic_allocation=True,
                tickers=[TickerConfig(symbol="AAPL", category="MOMENTUM", allocation_pct=100.0)],
            ),
        )

        # Mock fetcher
        mock_fetcher = Mock()
        mock_fetcher.get_stock_data.side_effect = [
            pd.DataFrame(),  # Empty on start date
            pd.DataFrame({"Close": [150.0]}, index=[pd.Timestamp("2023-01-08")]),  # Week later
        ]
        mock_fetcher.get_current_prices.return_value = {"AAPL": 155.0}

        # Update the allocator's fetcher
        mock_allocator.fetcher = mock_fetcher

        factory = DomainFactory(
            config=config, fetcher=mock_fetcher, allocator_manager=mock_allocator, logging_manager=mock_logging_manager
        )

        # Calculate quantities
        quantities = factory.allocator_manager.calculate_dynamic_quantities(config, config.portfolio.tickers)

        assert "AAPL" in quantities
        assert quantities["AAPL"] > 0

    def test_auto_allocation_with_no_category_ratios(self, mock_logging_manager):
        """Test auto-allocation without category ratios."""
        # PortfolioConfig validation prevents auto_allocate without category_ratios
        # So we test that the validation works
        with pytest.raises(ValueError, match="auto_allocate=True requires category_ratios"):
            StockulaConfig(
                portfolio=PortfolioConfig(
                    name="Test Portfolio",
                    initial_capital=100000,
                    auto_allocate=True,
                    tickers=[TickerConfig(symbol="AAPL", category="MOMENTUM")],
                )
            )

    def test_auto_allocation_with_fractional_shares(self, mock_allocator):
        """Test auto-allocation with fractional shares allowed."""
        config = StockulaConfig(
            portfolio=PortfolioConfig(
                name="Test Portfolio",
                initial_capital=100000,
                auto_allocate=True,
                allow_fractional_shares=True,
                category_ratios={"MOMENTUM": 0.6, "GROWTH": 0.4},
                capital_utilization_target=0.95,
                tickers=[
                    TickerConfig(symbol="AAPL", category="MOMENTUM"),
                    TickerConfig(symbol="GOOGL", category="GROWTH"),
                    TickerConfig(symbol="MSFT", category="GROWTH"),
                ],
            )
        )

        mock_fetcher = Mock()
        mock_fetcher.get_current_prices.return_value = {
            "AAPL": 150.0,
            "GOOGL": 120.0,
            "MSFT": 300.0,
        }

        # Update the allocator's fetcher
        mock_allocator.fetcher = mock_fetcher

        factory = DomainFactory(
            config=config, fetcher=mock_fetcher, allocator_manager=mock_allocator, logging_manager=mock_logging_manager
        )

        quantities = factory.allocator_manager.calculate_auto_allocation_quantities(config, config.portfolio.tickers)

        # Check that quantities are fractional
        assert "AAPL" in quantities
        assert "GOOGL" in quantities
        assert "MSFT" in quantities
        # With fractional shares, we should use most of the capital
        total_value = quantities["AAPL"] * 150.0 + quantities["GOOGL"] * 120.0 + quantities["MSFT"] * 300.0
        assert abs(total_value - 95000.0) < 100  # Close to 95% of 100k

    def test_auto_allocation_with_zero_ratio_category(self, mock_allocator, mock_logging_manager):
        """Test auto-allocation with zero ratio for a category."""
        config = StockulaConfig(
            portfolio=PortfolioConfig(
                name="Test Portfolio",
                initial_capital=100000,
                auto_allocate=True,
                category_ratios={"MOMENTUM": 1.0, "GROWTH": 0.0},
                capital_utilization_target=1.0,
                tickers=[
                    TickerConfig(symbol="AAPL", category="MOMENTUM"),
                    TickerConfig(symbol="GOOGL", category="GROWTH"),
                ],
            )
        )

        mock_fetcher = Mock()
        mock_fetcher.get_current_prices.return_value = {
            "AAPL": 150.0,
            "GOOGL": 120.0,
        }

        # Update the allocator's fetcher
        mock_allocator.fetcher = mock_fetcher

        factory = DomainFactory(
            config=config, fetcher=mock_fetcher, allocator_manager=mock_allocator, logging_manager=mock_logging_manager
        )

        quantities = factory.allocator_manager.calculate_auto_allocation_quantities(config, config.portfolio.tickers)

        # AAPL should get allocation, GOOGL should get 0
        assert quantities["AAPL"] > 0
        assert quantities["GOOGL"] == 0

    def test_create_portfolio_with_missing_info(self, mock_logging_manager):
        """Test portfolio creation when ticker info is missing."""
        config = StockulaConfig(
            portfolio=PortfolioConfig(
                name="Test Portfolio",
                initial_capital=100000,
                tickers=[TickerConfig(symbol="UNKNOWN", quantity=10)],
            )
        )

        mock_fetcher = Mock()
        # Return minimal info
        mock_fetcher.get_info.return_value = {"symbol": "UNKNOWN"}
        mock_fetcher.get_current_prices.return_value = {"UNKNOWN": 100.0}

        factory = DomainFactory(config=config, fetcher=mock_fetcher, logging_manager=mock_logging_manager)

        # Should create portfolio without error
        portfolio = factory.create_portfolio(config)
        assert portfolio is not None
        assert len(portfolio.assets) == 1

    def test_create_asset_with_invalid_category(self, mock_logging_manager):
        """Test creating asset with invalid category string."""
        config = StockulaConfig(
            portfolio=PortfolioConfig(
                name="Test Portfolio",
                initial_capital=100000,
                tickers=[TickerConfig(symbol="AAPL", quantity=10, category="INVALID_CATEGORY")],
            )
        )

        factory = DomainFactory(config=config, logging_manager=mock_logging_manager)

        # Should create asset without error (category will be None)
        ticker_config = config.portfolio.tickers[0]
        asset = factory._create_asset(ticker_config)

        # Asset should be created but category might be None
        assert asset is not None
        assert asset.ticker.symbol == "AAPL"
        assert asset.quantity == 10

    def test_calculate_dynamic_quantities_missing_price(self, mock_logging_manager):
        """Test dynamic quantity calculation when price is missing."""
        config = StockulaConfig(
            portfolio=PortfolioConfig(
                name="Test Portfolio",
                initial_capital=100000,
                dynamic_allocation=True,
                tickers=[TickerConfig(symbol="AAPL", allocation_pct=100.0)],
            )
        )

        mock_fetcher = Mock()
        mock_fetcher.get_current_prices.return_value = {}  # No prices returned

        mock_allocator_manager = Mock()
        mock_allocator_manager.calculate_dynamic_quantities.side_effect = ValueError("Could not fetch price for AAPL")

        factory = DomainFactory(
            config=config,
            fetcher=mock_fetcher,
            allocator_manager=mock_allocator_manager,
            logging_manager=mock_logging_manager,
        )

        with pytest.raises(ValueError, match="Could not fetch price"):
            factory.allocator_manager.calculate_dynamic_quantities(config, config.portfolio.tickers)

    def test_ticker_config_allows_no_allocation(self, mock_logging_manager):
        """Test that TickerConfig allows no allocation for backtest_optimized."""
        # This is now valid for backtest_optimized allocation
        ticker = TickerConfig(symbol="AAPL")
        assert ticker.symbol == "AAPL"
        assert ticker.quantity is None
        assert ticker.allocation_pct is None
        assert ticker.allocation_amount is None

    def test_auto_allocation_requires_category_for_all_tickers(self, mock_logging_manager):
        """Test that auto-allocation requires category for all tickers."""
        # When auto_allocate is True, ALL tickers must have category
        config = StockulaConfig(
            portfolio=PortfolioConfig(
                name="Test Portfolio",
                initial_capital=100000,
                auto_allocate=True,
                category_ratios={"MOMENTUM": 1.0},
                tickers=[
                    TickerConfig(symbol="AAPL", category="MOMENTUM"),
                    TickerConfig(
                        symbol="GOOGL", quantity=10
                    ),  # Has quantity but no category - invalid for auto-allocation
                ],
            )
        )

        mock_fetcher = Mock()
        mock_fetcher.get_current_prices.return_value = {"AAPL": 150.0, "GOOGL": 120.0}

        mock_allocator_manager = Mock()
        mock_allocator_manager.calculate_auto_allocation_quantities.side_effect = ValueError(
            "Ticker GOOGL must have category specified for auto-allocation"
        )

        factory = DomainFactory(
            config=config,
            fetcher=mock_fetcher,
            allocator_manager=mock_allocator_manager,
            logging_manager=mock_logging_manager,
        )

        with pytest.raises(ValueError, match="must have category specified for auto-allocation"):
            factory.create_portfolio(config)

    def test_get_all_tickers_from_factory(self, mock_logging_manager):
        """Test getting all registered tickers from factory."""
        factory = DomainFactory(logging_manager=mock_logging_manager)

        # Create some tickers with quantity
        ticker_config1 = TickerConfig(symbol="AAPL", sector="Technology", quantity=10)
        ticker_config2 = TickerConfig(symbol="GOOGL", sector="Technology", quantity=20)

        factory._create_ticker(ticker_config1)
        factory._create_ticker(ticker_config2)

        all_tickers = factory.get_all_tickers()

        assert len(all_tickers) == 2
        symbols = [t.symbol for t in all_tickers]
        assert "AAPL" in symbols
        assert "GOOGL" in symbols



"""Settings management using pydantic-settings."""

from pathlib import Path
from typing import Any, cast

import yaml  # type: ignore[import-untyped]
from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict

from .models import MACDConfig, RSIConfig, SMACrossConfig, StockulaConfig, StrategyConfig


class Settings(BaseSettings):
    """Application settings with environment variable support."""

    model_config = SettingsConfigDict(
        env_prefix="STOCKULA_",
        env_file=".env",
        env_file_encoding="utf-8",
        case_sensitive=False,
    )

    config_file: str | None = Field(default=None, description="Path to YAML configuration file")
    debug: bool = Field(default=False, description="Enable debug mode")
    log_level: str = Field(default="INFO", description="Logging level")


def load_yaml_config(config_path: str | Path) -> dict[str, Any]:
    """Load configuration from YAML file.

    Args:
        config_path: Path to YAML configuration file

    Returns:
        Dictionary containing configuration data
    """
    config_path = Path(config_path)
    if not config_path.exists():
        raise FileNotFoundError(f"Configuration file not found: {config_path}")

    with open(config_path) as f:
        return cast(dict[str, Any], yaml.safe_load(f))


def parse_strategy_config(strategy_dict: dict[str, Any]) -> StrategyConfig:
    """Parse strategy configuration from dictionary.

    Args:
        strategy_dict: Dictionary containing strategy configuration

    Returns:
        StrategyConfig instance
    """
    strategy_name = strategy_dict.get("name", "").lower()
    params = strategy_dict.get("parameters", {})

    # Map strategy names to specific config classes
    if strategy_name == "smacross":
        params_model: SMACrossConfig | RSIConfig | MACDConfig = SMACrossConfig(**params)
        params = params_model.model_dump()
    elif strategy_name == "rsi":
        params_model = RSIConfig(**params)
        params = params_model.model_dump()
    elif strategy_name == "macd":
        params_model = MACDConfig(**params)
        params = params_model.model_dump()

    return StrategyConfig(name=strategy_dict["name"], parameters=params)


def load_config(config_path: str | Path | None = None) -> StockulaConfig:
    """Load configuration from file or environment.

    Args:
        config_path: Optional path to configuration file.
                    If not provided, checks for:
                    1. STOCKULA_CONFIG_FILE env var
                    2. .stockula.yaml in current directory
                    3. .stockula.yml in current directory
                    4. stockula.yaml in current directory
                    5. stockula.yml in current directory

    Returns:
        StockulaConfig instance
    """
    settings = Settings()

    # Determine config file path
    if config_path is None:
        config_path = settings.config_file

        # If no env var, check for default files
        if config_path is None:
            default_files = [
                ".stockula.yaml",
                ".stockula.yml",
                "stockula.yaml",
                "stockula.yml",
            ]
            for filename in default_files:
                if Path(filename).exists():
                    config_path = filename
                    break

    # If no config file specified or found, return default configuration
    if config_path is None:
        return StockulaConfig()

    # Load YAML configuration
    config_data = load_yaml_config(config_path)

    # Parse strategies if present
    if "backtest" in config_data and "strategies" in config_data["backtest"]:
        strategies = []
        for strategy_dict in config_data["backtest"]["strategies"]:
            strategies.append(parse_strategy_config(strategy_dict))
        config_data["backtest"]["strategies"] = strategies

    # Create and return configuration object
    return StockulaConfig(**config_data)


def save_config(config: StockulaConfig, config_path: str | Path) -> None:
    """Save configuration to YAML file.

    Args:
        config: Configuration object to save
        config_path: Path to save configuration file
    """
    config_path = Path(config_path)
    config_path.parent.mkdir(parents=True, exist_ok=True)

    # Convert to dictionary for YAML serialization
    config_dict = config.model_dump(mode="json")

    with open(config_path, "w") as f:
        yaml.dump(config_dict, f, default_flow_style=False, sort_keys=False)
